# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations

from typing import Set
from PySide6 import QtWidgets
import numpy as np
from VeraGrid.Gui.GridReduce.grid_reduce_gui import Ui_ReduceDialog
from VeraGrid.Gui.general_dialogues import LogsDialogue
from VeraGrid.Gui.messages import yes_no_question, warning_msg
from VeraGrid.Gui.gui_functions import get_list_model
from VeraGrid.Session.session import SimulationSession
from VeraGridEngine.Devices.Substation.bus import Bus
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.Topology.GridReduction.di_shi_grid_reduction import di_shi_reduction
from VeraGridEngine.Topology.GridReduction.ptdf_grid_reduction import ptdf_reduction
from VeraGridEngine.Topology.GridReduction.ward_equivalents import ward_standard_reduction
from VeraGridEngine.Simulations.LinearFactors.linear_analysis import LinearAnalysis, LinearAnalysisTs
from VeraGridEngine.Compilers.circuit_to_data import compile_numerical_circuit_at
from VeraGridEngine.basic_structures import Logger
from VeraGridEngine.enumerations import GridReductionMethod


class GridReduceDialogue(QtWidgets.QDialog):
    """
    GridMergeDialogue
    """

    def __init__(self, grid: MultiCircuit, session: SimulationSession, selected_buses_set: Set[Bus]):
        """
        GridMergeDialogue
        :param grid: MultiCircuit instance
        :param session: SimulationSession instance to query the power flow
        :param selected_buses_set: Set of buses (Bus objects) to reduce
        """
        QtWidgets.QDialog.__init__(self)
        self.ui = Ui_ReduceDialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Grid reduction')

        self.logger = Logger()
        self.logs_dialogue: LogsDialogue | None = None

        self.ui.listView.setModel(get_list_model(list(selected_buses_set)))

        methods = [GridReductionMethod.DiShi, GridReductionMethod.PTDF, GridReductionMethod.Ward]
        self.methods_dict = {m.value: m for m in methods}
        self.ui.methodComboBox.setModel(get_list_model([m.value for m in methods]))

        self._grid: MultiCircuit = grid
        self._session: SimulationSession = session
        self._selected_buses_set: Set[Bus] = selected_buses_set

        self.ui.reduceButton.clicked.connect(self.reduce_grid)

    def reduce_grid(self):
        """
        The elements of the grid will be added with new idtags.
        This is useful in the case you want to compose a new grid from grids that are the same.
        :return:
        """
        if len(self._selected_buses_set):

            method: GridReductionMethod = self.methods_dict[self.ui.methodComboBox.currentText()]

            ok = yes_no_question(
                text="This will delete the selected buses and reintroduce their influence"
                     "using the Ward equivalent. This cannot be undone and it is dangerous if you don't know"
                     "what you are doing \nAre you sure?",
                title="Grid reduction?")

            if ok:

                # convert the set of buses to bus indices
                reduction_bus_indices = np.array([self._grid.buses.index(b) for b in self._selected_buses_set],
                                                 dtype=int)

                if method == GridReductionMethod.DiShi:

                    # get the previous power flow
                    _, pf_res = self._session.power_flow

                    if pf_res is None:
                        warning_msg("Run a power flow first! or select another method", "Grid reduction")
                        return

                    # NOTE: self._grid gets reduced in-place
                    grid_reduced, logger = di_shi_reduction(
                        grid=self._grid,
                        reduction_bus_indices=reduction_bus_indices,
                        V0=pf_res.voltage
                    )

                elif method == GridReductionMethod.Ward:

                    # get the previous power flow
                    _, pf_res = self._session.power_flow

                    if pf_res is None:
                        warning_msg("Run a power flow first! or select another method", "Grid reduction")
                        return

                    # NOTE: self._grid gets reduced in-place
                    grid_reduced, logger = ward_standard_reduction(
                        grid=self._grid,
                        reduction_bus_indices=reduction_bus_indices,
                        V0=pf_res.voltage
                    )

                elif method == GridReductionMethod.PTDF:
                    # get the previous power flow
                    nc = compile_numerical_circuit_at(circuit=self._grid, t_idx=None)
                    lin = LinearAnalysis(nc=nc)

                    if self._grid.has_time_series:
                        lin_ts = LinearAnalysisTs(grid=self._grid)
                    else:
                        lin_ts = None

                    # NOTE: self._grid gets reduced in-place
                    grid_reduced, logger = ptdf_reduction(
                        grid=self._grid,
                        reduction_bus_indices=reduction_bus_indices,
                        PTDF=lin.PTDF,
                        lin_ts=lin_ts
                    )
                else:
                    raise NotImplementedError("Reduction method not supported")

                if logger.has_logs():
                    self.logs_dialogue = LogsDialogue(name="Import profiles", logger=logger)
                    self.logs_dialogue.exec()
            else:
                pass  # not ok
        else:
            warning_msg("No reduction happened", "Grid reduction")

        # exit
        self.close()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'load_designer_ui.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QDoubleSpinBox, QFrame,
    QHBoxLayout, QHeaderView, QLabel, QPushButton,
    QSizePolicy, QSpacerItem, QSplitter, QTableView,
    QTimeEdit, QToolBox, QVBoxLayout, QWidget)

from VeraGrid.Gui.Widgets.matplotlibwidget import MatplotlibWidget
from .icons_rc import *

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(941, 590)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(1, 1, 1, 1)
        self.splitter_3 = QSplitter(Dialog)
        self.splitter_3.setObjectName(u"splitter_3")
        self.splitter_3.setOrientation(Qt.Horizontal)
        self.frame_8 = QFrame(self.splitter_3)
        self.frame_8.setObjectName(u"frame_8")
        self.frame_8.setFrameShape(QFrame.NoFrame)
        self.frame_8.setFrameShadow(QFrame.Raised)
        self.verticalLayout_7 = QVBoxLayout(self.frame_8)
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.verticalLayout_7.setContentsMargins(-1, 9, -1, -1)
        self.toolBox = QToolBox(self.frame_8)
        self.toolBox.setObjectName(u"toolBox")
        self.page_3 = QWidget()
        self.page_3.setObjectName(u"page_3")
        self.page_3.setGeometry(QRect(0, 0, 848, 352))
        self.verticalLayout_3 = QVBoxLayout(self.page_3)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, -1, 0, -1)
        self.frame_2 = QFrame(self.page_3)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setMinimumSize(QSize(300, 0))
        self.frame_2.setFrameShape(QFrame.StyledPanel)
        self.frame_2.setFrameShadow(QFrame.Raised)
        self.verticalLayout_4 = QVBoxLayout(self.frame_2)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.frame_3 = QFrame(self.frame_2)
        self.frame_3.setObjectName(u"frame_3")
        self.frame_3.setFrameShape(QFrame.NoFrame)
        self.frame_3.setFrameShadow(QFrame.Plain)
        self.horizontalLayout_3 = QHBoxLayout(self.frame_3)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.label = QLabel(self.frame_3)
        self.label.setObjectName(u"label")
        self.label.setMaximumSize(QSize(160, 16777215))

        self.horizontalLayout_3.addWidget(self.label)

        self.night_valley_timeEdit = QTimeEdit(self.frame_3)
        self.night_valley_timeEdit.setObjectName(u"night_valley_timeEdit")
        self.night_valley_timeEdit.setTime(QTime(3, 0, 0))

        self.horizontalLayout_3.addWidget(self.night_valley_timeEdit)

        self.night_valley_doubleSpinBox = QDoubleSpinBox(self.frame_3)
        self.night_valley_doubleSpinBox.setObjectName(u"night_valley_doubleSpinBox")
        self.night_valley_doubleSpinBox.setDecimals(4)
        self.night_valley_doubleSpinBox.setMaximum(9999999.000000000000000)
        self.night_valley_doubleSpinBox.setValue(4.000000000000000)

        self.horizontalLayout_3.addWidget(self.night_valley_doubleSpinBox)


        self.verticalLayout_4.addWidget(self.frame_3)

        self.frame_4 = QFrame(self.frame_2)
        self.frame_4.setObjectName(u"frame_4")
        self.frame_4.setFrameShape(QFrame.NoFrame)
        self.frame_4.setFrameShadow(QFrame.Plain)
        self.horizontalLayout_4 = QHBoxLayout(self.frame_4)
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.label_2 = QLabel(self.frame_4)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setMaximumSize(QSize(160, 16777215))

        self.horizontalLayout_4.addWidget(self.label_2)

        self.morning_peak_timeEdit = QTimeEdit(self.frame_4)
        self.morning_peak_timeEdit.setObjectName(u"morning_peak_timeEdit")
        self.morning_peak_timeEdit.setTime(QTime(10, 0, 0))

        self.horizontalLayout_4.addWidget(self.morning_peak_timeEdit)

        self.morning_peak_doubleSpinBox = QDoubleSpinBox(self.frame_4)
        self.morning_peak_doubleSpinBox.setObjectName(u"morning_peak_doubleSpinBox")
        self.morning_peak_doubleSpinBox.setDecimals(4)
        self.morning_peak_doubleSpinBox.setMaximum(9999999.000000000000000)
        self.morning_peak_doubleSpinBox.setValue(12.000000000000000)

        self.horizontalLayout_4.addWidget(self.morning_peak_doubleSpinBox)


        self.verticalLayout_4.addWidget(self.frame_4)

        self.frame_5 = QFrame(self.frame_2)
        self.frame_5.setObjectName(u"frame_5")
        self.frame_5.setFrameShape(QFrame.NoFrame)
        self.frame_5.setFrameShadow(QFrame.Plain)
        self.horizontalLayout_6 = QHBoxLayout(self.frame_5)
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.label_4 = QLabel(self.frame_5)
        self.label_4.setObjectName(u"label_4")
        self.label_4.setMaximumSize(QSize(160, 16777215))

        self.horizontalLayout_6.addWidget(self.label_4)

        self.afternoon_valley_timeEdit = QTimeEdit(self.frame_5)
        self.afternoon_valley_timeEdit.setObjectName(u"afternoon_valley_timeEdit")
        self.afternoon_valley_timeEdit.setTime(QTime(15, 0, 0))

        self.horizontalLayout_6.addWidget(self.afternoon_valley_timeEdit)

        self.afternoon_valley_doubleSpinBox = QDoubleSpinBox(self.frame_5)
        self.afternoon_valley_doubleSpinBox.setObjectName(u"afternoon_valley_doubleSpinBox")
        self.afternoon_valley_doubleSpinBox.setDecimals(4)
        self.afternoon_valley_doubleSpinBox.setMaximum(9999999.000000000000000)
        self.afternoon_valley_doubleSpinBox.setValue(8.000000000000000)

        self.horizontalLayout_6.addWidget(self.afternoon_valley_doubleSpinBox)


        self.verticalLayout_4.addWidget(self.frame_5)

        self.frame_6 = QFrame(self.frame_2)
        self.frame_6.setObjectName(u"frame_6")
        self.frame_6.setFrameShape(QFrame.NoFrame)
        self.frame_6.setFrameShadow(QFrame.Plain)
        self.horizontalLayout_7 = QHBoxLayout(self.frame_6)
        self.horizontalLayout_7.setObjectName(u"horizontalLayout_7")
        self.label_5 = QLabel(self.frame_6)
        self.label_5.setObjectName(u"label_5")
        self.label_5.setMaximumSize(QSize(160, 16777215))

        self.horizontalLayout_7.addWidget(self.label_5)

        self.evening_peak_timeEdit = QTimeEdit(self.frame_6)
        self.evening_peak_timeEdit.setObjectName(u"evening_peak_timeEdit")
        self.evening_peak_timeEdit.setTime(QTime(20, 0, 0))

        self.horizontalLayout_7.addWidget(self.evening_peak_timeEdit)

        self.evening_peak_doubleSpinBox = QDoubleSpinBox(self.frame_6)
        self.evening_peak_doubleSpinBox.setObjectName(u"evening_peak_doubleSpinBox")
        self.evening_peak_doubleSpinBox.setDecimals(4)
        self.evening_peak_doubleSpinBox.setMaximum(9999999.000000000000000)
        self.evening_peak_doubleSpinBox.setValue(16.000000000000000)

        self.horizontalLayout_7.addWidget(self.evening_peak_doubleSpinBox)


        self.verticalLayout_4.addWidget(self.frame_6)

        self.frame_7 = QFrame(self.frame_2)
        self.frame_7.setObjectName(u"frame_7")
        self.frame_7.setFrameShape(QFrame.NoFrame)
        self.frame_7.setFrameShadow(QFrame.Plain)
        self.horizontalLayout_8 = QHBoxLayout(self.frame_7)
        self.horizontalLayout_8.setObjectName(u"horizontalLayout_8")
        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_8.addItem(self.horizontalSpacer_2)

        self.draw_by_peak_pushButton = QPushButton(self.frame_7)
        self.draw_by_peak_pushButton.setObjectName(u"draw_by_peak_pushButton")
        icon = QIcon()
        icon.addFile(u":/Icons/icons/gear.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.draw_by_peak_pushButton.setIcon(icon)

        self.horizontalLayout_8.addWidget(self.draw_by_peak_pushButton)


        self.verticalLayout_4.addWidget(self.frame_7)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_4.addItem(self.verticalSpacer)


        self.verticalLayout_3.addWidget(self.frame_2)

        self.toolBox.addItem(self.page_3, u"Definition by peak points")
        self.page = QWidget()
        self.page.setObjectName(u"page")
        self.page.setGeometry(QRect(0, 0, 848, 352))
        self.verticalLayout_2 = QVBoxLayout(self.page)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.tableView = QTableView(self.page)
        self.tableView.setObjectName(u"tableView")

        self.verticalLayout_2.addWidget(self.tableView)

        self.frame = QFrame(self.page)
        self.frame.setObjectName(u"frame")
        self.frame.setMaximumSize(QSize(16777215, 40))
        self.frame.setFrameShape(QFrame.NoFrame)
        self.frame.setFrameShadow(QFrame.Raised)
        self.horizontalLayout_2 = QHBoxLayout(self.frame)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer)

        self.draw_by_points_pushButton = QPushButton(self.frame)
        self.draw_by_points_pushButton.setObjectName(u"draw_by_points_pushButton")
        self.draw_by_points_pushButton.setIcon(icon)

        self.horizontalLayout_2.addWidget(self.draw_by_points_pushButton)


        self.verticalLayout_2.addWidget(self.frame)

        self.toolBox.addItem(self.page, u"Definition by data points")

        self.verticalLayout_7.addWidget(self.toolBox)

        self.frame_9 = QFrame(self.frame_8)
        self.frame_9.setObjectName(u"frame_9")
        self.frame_9.setMinimumSize(QSize(0, 150))
        self.frame_9.setFrameShape(QFrame.StyledPanel)
        self.frame_9.setFrameShadow(QFrame.Raised)
        self.horizontalLayout_9 = QHBoxLayout(self.frame_9)
        self.horizontalLayout_9.setObjectName(u"horizontalLayout_9")

        self.verticalLayout_7.addWidget(self.frame_9)

        self.splitter_3.addWidget(self.frame_8)
        self.PlotFrame = QFrame(self.splitter_3)
        self.PlotFrame.setObjectName(u"PlotFrame")
        self.PlotFrame.setFrameShape(QFrame.NoFrame)
        self.PlotFrame.setFrameShadow(QFrame.Raised)
        self.horizontalLayout = QHBoxLayout(self.PlotFrame)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.plotwidget = MatplotlibWidget(self.PlotFrame)
        self.plotwidget.setObjectName(u"plotwidget")

        self.horizontalLayout.addWidget(self.plotwidget)

        self.splitter_3.addWidget(self.PlotFrame)

        self.verticalLayout.addWidget(self.splitter_3)


        self.retranslateUi(Dialog)

        self.toolBox.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Dialog", None))
        self.label.setText(QCoreApplication.translate("Dialog", u"Night valley", None))
        self.night_valley_doubleSpinBox.setSuffix(QCoreApplication.translate("Dialog", u" MW", None))
        self.label_2.setText(QCoreApplication.translate("Dialog", u"Morning Peak", None))
        self.morning_peak_doubleSpinBox.setSuffix(QCoreApplication.translate("Dialog", u" MW", None))
        self.label_4.setText(QCoreApplication.translate("Dialog", u"Afternoon valley", None))
        self.afternoon_valley_doubleSpinBox.setSuffix(QCoreApplication.translate("Dialog", u" MW", None))
        self.label_5.setText(QCoreApplication.translate("Dialog", u"Evening Peak", None))
        self.evening_peak_doubleSpinBox.setSuffix(QCoreApplication.translate("Dialog", u" MW", None))
        self.draw_by_peak_pushButton.setText("")
        self.toolBox.setItemText(self.toolBox.indexOf(self.page_3), QCoreApplication.translate("Dialog", u"Definition by peak points", None))
        self.draw_by_points_pushButton.setText("")
        self.toolBox.setItemText(self.toolBox.indexOf(self.page), QCoreApplication.translate("Dialog", u"Definition by data points", None))
    # retranslateUi




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
from typing import TYPE_CHECKING
from VeraGrid.Gui.Diagrams.generic_graphics import Square
from VeraGridEngine.Devices.Injections.current_injection import CurrentInjection
from VeraGrid.Gui.Diagrams.SchematicWidget.Injections.injections_template_graphics import InjectionTemplateGraphicItem


if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget


class CurrentInjectionGraphicItem(InjectionTemplateGraphicItem):
    """
    ExternalGrid graphic item
    """

    def __init__(self, parent, api_obj: CurrentInjection, editor: SchematicWidget):
        """

        :param parent:
        :param api_obj:
        :param editor:
        """
        InjectionTemplateGraphicItem.__init__(self,
                                              parent=parent,
                                              api_obj=api_obj,
                                              editor=editor,
                                              device_type_name='Current injection',
                                              w=40,
                                              h=40)
        self.set_glyph(glyph=Square(self, 40, 40, "I", self.update_nexus))

    @property
    def api_object(self) -> CurrentInjection:
        return self._api_object




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
import sys
import numpy as np
from typing import Union, TYPE_CHECKING
from PySide6.QtCore import Qt, QLineF, QPointF, QRectF
from PySide6.QtGui import QPen, QCursor, QPixmap, QBrush, QColor, QTransform, QPolygonF
from PySide6.QtWidgets import (QGraphicsLineItem, QGraphicsRectItem, QGraphicsPolygonItem,
                               QGraphicsEllipseItem, QGraphicsSceneMouseEvent, QGraphicsTextItem)
from VeraGrid.Gui.Diagrams.generic_graphics import ACTIVE, DEACTIVATED, OTHER, GenericDiagramWidget, TRANSPARENT, WHITE
from VeraGrid.Gui.Diagrams.SchematicWidget.terminal_item import BarTerminalItem, RoundTerminalItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Substation.bus_graphics import BusGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Fluid.fluid_node_graphics import FluidNodeGraphicItem
from VeraGrid.Gui.messages import yes_no_question

from VeraGridEngine.Devices.Substation.bus import Bus
from VeraGridEngine.Devices.Substation.busbar import BusBar
from VeraGridEngine.Devices.Branches.line import Line
from VeraGridEngine.Devices.Branches.transformer import Transformer2W
from VeraGridEngine.Devices.Branches.winding import Winding
from VeraGridEngine.Devices.Branches.vsc import VSC
from VeraGridEngine.Devices.Branches.upfc import UPFC
from VeraGridEngine.Devices.Branches.switch import Switch
from VeraGridEngine.Devices.Branches.dc_line import DcLine
from VeraGridEngine.Devices.Branches.series_reactance import SeriesReactance
from VeraGridEngine.Devices.Branches.hvdc_line import HvdcLine
from VeraGridEngine.Devices.Fluid.fluid_node import FluidNode
from VeraGridEngine.Devices.Fluid.fluid_path import FluidPath
from VeraGridEngine.Devices.types import BRANCH_TYPES
from VeraGridEngine.enumerations import SwitchGraphicType

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget
    from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.transformer3w_graphics import Transformer3WGraphicItem
    from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.vsc_graphics_3term import VscGraphicItem3Term

NODE_GRAPHIC = BusGraphicItem | FluidNodeGraphicItem


class TransformerSymbol(QGraphicsRectItem):
    """
    TransformerSymbol
    """

    def __init__(self, parent, pen_width: int, h=80, w=80):
        QGraphicsRectItem.__init__(self, parent=parent)

        d = w / 2

        self.parent = parent

        self.width = pen_width
        self.pen_width = pen_width
        self.color = ACTIVE['color']
        self.style = ACTIVE['style']

        self.setPen(QPen(TRANSPARENT))
        self.setRect(QRectF(0, 0, w, h))

        self.c0 = QGraphicsEllipseItem(0, 0, d, d, parent=self)
        self.c1 = QGraphicsEllipseItem(0, 0, d, d, parent=self)
        self.c2 = QGraphicsEllipseItem(0, 0, d, d, parent=self)

        self.c0.setPen(QPen(TRANSPARENT, self.width, self.style))
        self.c2.setPen(QPen(self.color, self.width, self.style))
        self.c1.setPen(QPen(self.color, self.width, self.style))

        self.c0.setBrush(QBrush(WHITE))
        self.c2.setBrush(QBrush(WHITE))

        self.c0.setPos(w * 0.35 - d / 2, h * 0.5 - d / 2)
        self.c1.setPos(w * 0.35 - d / 2, h * 0.5 - d / 2)
        self.c2.setPos(w * 0.65 - d / 2, h * 0.5 - d / 2)

        self.c0.setZValue(0)
        self.c1.setZValue(2)
        self.c2.setZValue(1)

    def set_colour(self, color: QColor, w, style: Qt.PenStyle):
        """
        Set color and style
        :param color: QColor instance
        :param w: width
        :param style: PenStyle instance
        :return:
        """
        self.c2.setPen(QPen(color, w, style))
        self.c1.setPen(QPen(color, w, style))

    def set_pen(self, pen: QPen):
        """

        :param pen:
        :return:
        """
        self.setPen(pen)
        self.c1.setPen(pen)
        self.c2.setPen(pen)

    def setToolTipText(self, toolTip: str):
        """
        Set branch tool tip text
        Args:
            toolTip: text
        """
        self.setToolTip(toolTip)
        self.c0.setToolTip(toolTip)
        self.c1.setToolTip(toolTip)
        self.c2.setToolTip(toolTip)

    def redraw(self):
        """

        :return:
        """
        h = self.parent.pos2.y() - self.parent.pos1.y()
        b = self.parent.pos2.x() - self.parent.pos1.x()
        ang = np.arctan2(h, b)
        h2 = self.rect().height() / 2.0
        w2 = self.rect().width() / 2.0
        a = h2 * np.cos(ang) - w2 * np.sin(ang)
        b = w2 * np.sin(ang) + h2 * np.cos(ang)

        center = (self.parent.pos1 + self.parent.pos2) * 0.5 - QPointF(a, b)

        transform = QTransform()
        transform.translate(center.x(), center.y())
        transform.rotate(np.rad2deg(ang))
        self.setTransform(transform)


class VscSymbol(QGraphicsRectItem):
    """
    VscSymbol
    """

    def __init__(self, parent, pen_width, h=48, w=48, icon_route=":/Icons/icons/vsc.svg"):
        QGraphicsRectItem.__init__(self, parent=parent)

        self.parent = parent

        self.width = pen_width
        self.pen_width = pen_width
        self.color = ACTIVE['color']
        self.style = ACTIVE['style']

        self.setPen(QPen(TRANSPARENT))
        self.setRect(QRectF(0, 0, w, h))

        graphic = QGraphicsRectItem(QRectF(0, 0, w, h), parent=self)
        graphic.setBrush(QBrush(QPixmap(icon_route)))
        graphic.setPen(QPen(TRANSPARENT, self.width, self.style))

    def set_colour(self, color: QColor, w, style: Qt.PenStyle):
        """
        Set color and style
        :param color: QColor instance
        :param w: width
        :param style: PenStyle instance
        :return:
        """
        self.setBrush(color)
        self.setPen(QPen(color, w, style))

    def set_pen(self, pen: QPen):
        """

        :param pen:
        :return:
        """
        self.setPen(pen)

    def setToolTipText(self, toolTip: str):
        """
        Set branch tool tip text
        Args:
            toolTip: text
        """
        self.setToolTip(toolTip)

    def redraw(self):
        """

        :return:
        """
        h = self.parent.pos2.y() - self.parent.pos1.y()
        b = self.parent.pos2.x() - self.parent.pos1.x()
        ang = np.arctan2(h, b)
        h2 = self.rect().height() / 2.0
        w2 = self.rect().width() / 2.0
        a = h2 * np.cos(ang) - w2 * np.sin(ang)
        b = w2 * np.sin(ang) + h2 * np.cos(ang)

        center = (self.parent.pos1 + self.parent.pos2) * 0.5 - QPointF(a, b)

        transform = QTransform()
        transform.translate(center.x(), center.y())
        transform.rotate(np.rad2deg(ang))
        self.setTransform(transform)


class UpfcSymbol(VscSymbol):
    """
    UpfcSymbol
    """

    def __init__(self, parent, pen_width, h=48, w=48):
        VscSymbol.__init__(self, parent=parent, pen_width=pen_width, h=h, w=w, icon_route=":/Icons/icons/upfc.svg")


class SeriesReactanceSymbol(VscSymbol):
    """
    UpfcSymbol
    """

    def __init__(self, parent, pen_width, h=30, w=30):
        VscSymbol.__init__(self, parent=parent, pen_width=pen_width, h=h, w=w,
                           icon_route=":/Icons/icons/reactance.svg")


class SwitchSymbol(VscSymbol):
    """
    SwitchSymbol
    """

    def __init__(self, parent, pen_width, h=30, w=30):
        VscSymbol.__init__(self, parent=parent, pen_width=pen_width, h=h, w=w,
                           icon_route=":/Icons/icons/switch.svg")


class DisconnectorSymbol(VscSymbol):
    """
    DisconnectorSymbol
    """

    def __init__(self, parent, pen_width, h=30, w=30):
        VscSymbol.__init__(self, parent=parent, pen_width=pen_width, h=h, w=w,
                           icon_route=":/Icons/icons/disconnector.svg")


class HvdcSymbol(QGraphicsRectItem):
    """
    HvdcSymbol
    """

    def __init__(self, parent, pen_width, h=30, w=30):
        QGraphicsRectItem.__init__(self, parent=parent)

        w2 = int(w / 2)
        self.parent = parent

        self.width = pen_width
        self.pen_width = pen_width
        self.color = ACTIVE['color']
        self.style = ACTIVE['style']

        self.setPen(QPen(TRANSPARENT))
        self.setRect(QRectF(0, 0, w, h))

        offset = 3
        t_points = QPolygonF()
        t_points.append(QPointF(0, offset))
        t_points.append(QPointF(w - offset, w2))
        t_points.append(QPointF(0, w - offset))

        triangle = QGraphicsPolygonItem(self)
        triangle.setPolygon(t_points)
        triangle.setPen(QPen(WHITE))
        triangle.setBrush(QBrush(WHITE))

        line = QGraphicsRectItem(QRectF(h - offset, offset, offset, w - 2 * offset), parent=self)
        line.setPen(QPen(WHITE))
        line.setBrush(QBrush(WHITE))

    def set_colour(self, color: QColor, w, style: Qt.PenStyle):
        """
        Set color and style
        :param color: QColor instance
        :param w: width
        :param style: PenStyle instance
        :return:
        """
        self.setBrush(color)
        self.setPen(QPen(color, w, style))

    def set_pen(self, pen: QPen):
        """

        :param pen:
        :return:
        """
        self.setPen(pen)

    def setToolTipText(self, toolTip: str):
        """
        Set branch tool tip text
        Args:
            toolTip: text
        """
        self.setToolTip(toolTip)

    def redraw(self):
        """
        Redraw the HVDC symbol
        """
        h = self.parent.pos2.y() - self.parent.pos1.y()
        b = self.parent.pos2.x() - self.parent.pos1.x()
        ang = np.arctan2(h, b)
        h2 = self.rect().height() / 2.0
        w2 = self.rect().width() / 2.0
        a = h2 * np.cos(ang) - w2 * np.sin(ang)
        b = w2 * np.sin(ang) + h2 * np.cos(ang)

        center = (self.parent.pos1 + self.parent.pos2) * 0.5 - QPointF(a, b)

        transform = QTransform()
        transform.translate(center.x(), center.y())
        transform.rotate(np.rad2deg(ang))
        self.setTransform(transform)


class ArrowHead(QGraphicsPolygonItem):
    """
    This is the arrow object
    """

    def __init__(self,
                 parent: QGraphicsLineItem,
                 arrow_size: float,
                 position: float = 0.9,
                 under: bool = False,
                 backwards: bool = False,
                 separation: int = 5,
                 show_text: bool = True,
                 text_scale: float = 1.0):
        """
        Constructor
        :param parent: Parent line
        :param arrow_size: Size of the arrow
        :param position: proportion of the line where to locate the arrow
        :param under: Is it under?
        :param backwards: Is it backwards?
        :param separation: Separation
        :param show_text: Show the label?
        :param text_scale: Text scale
        """
        QGraphicsPolygonItem.__init__(self, parent=parent)

        self.parent: QGraphicsLineItem = parent
        self.arrow_size: float = arrow_size
        self.position: float = position
        self.under: bool = under
        self.backwards: float = backwards
        self.sep = separation

        self.label = QGraphicsTextItem(self)
        self.label.setPlainText("")
        self.show_text = show_text

        self.setPen(Qt.PenStyle.NoPen)

        # self.setFlag(self.GraphicsItemFlag.ItemIgnoresTransformations, True)

    def setScale(self, scale):
        """

        :param scale:
        :return:
        """
        super().setScale(scale)
        self.label.setScale(scale)

    def set_size(self, size: float):
        self.arrow_size: float = size
        self.redraw()

    def set_colour(self, color: QColor):
        """
        Set color and style
        :param color: QColor instance
        """
        self.setBrush(color)
        self.label.setDefaultTextColor(color)
        # self.label.setScale(w)

    def set_value(self, value: float, redraw=True, backwards=False, name="", units="", format_str="{:10.2f}",
                  draw_label: bool = True, visibility_filter_value=1e-4):
        """
        Set the sign with a value
        :param value: any real value
        :param redraw: redraw after the sign update
        :param backwards: draw backwards
        :param name: name of the displayed magnitude (i.e. Pf)
        :param units: the units of the displayed magnitude (i.e MW)
        :param format_str: the formatting string of the displayed magnitude
        :param visibility_filter_value: threshold to determine if to show this widget
        :param draw_label: Draw label
        """
        self.setVisible(bool(abs(value) > visibility_filter_value))
        self.backwards = backwards

        self.label.setVisible(draw_label)
        if draw_label:
            x = format_str.format(value)
            msg = f'{name}:{x} {units}'
            self.label.setPlainText(msg)
            self.setToolTip(msg)

        if redraw:
            self.redraw()

    def redraw(self) -> None:
        """
        Redraw the arrow
        """
        line = self.parent.line()

        # the angle is added 180º if the sign is negative
        angle = - line.angle()
        base_pt = line.p1() + (line.p2() - line.p1()) * self.position

        p1 = -self.arrow_size if self.backwards else self.arrow_size
        p2 = -self.arrow_size if self.under else self.arrow_size
        arrow_p1 = base_pt - QTransform().rotate(angle).map(QPointF(p1, 0))
        arrow_p2 = base_pt - QTransform().rotate(angle).map(QPointF(p1, p2))
        arrow_polygon = QPolygonF([base_pt, arrow_p1, arrow_p2])

        self.setPolygon(arrow_polygon)

        if self.show_text:
            # if 90 < line.angle() <= 270:
            #     label_p = base_pt - QTransform().rotate(-angle).map(QPointF(20, -10 if self.under else 35))
            #     self.label.setPos(label_p)
            #     self.label.setRotation(-angle)
            # else:
            #     label_p = base_pt - QTransform().rotate(angle).map(QPointF(20, -10 if self.under else 35))
            #     self.label.setPos(label_p)
            #     self.label.setRotation(angle)

            if 90 < line.angle() <= 270:
                label_p = base_pt - QTransform().rotate(angle).map(QPointF(-20, -35 if self.under else +10))
                self.label.setRotation(angle + 180)
            else:
                label_p = base_pt - QTransform().rotate(angle).map(QPointF(40, -10 if self.under else 35))
                self.label.setRotation(angle)

            self.label.setPos(label_p)


class LineGraphicTemplateItem(GenericDiagramWidget, QGraphicsLineItem):
    """
    LineGraphicItem
    """

    def __init__(self,
                 from_port: Union[BarTerminalItem, RoundTerminalItem],
                 to_port: Union[BarTerminalItem, RoundTerminalItem, None],
                 editor: SchematicWidget,
                 width=5,
                 api_object: Union[Line, Transformer2W, VSC, UPFC, HvdcLine, DcLine, FluidPath, None] = None,
                 arrow_size=10,
                 draw_labels: bool = True,
                 arrow_f_pos=0.2,
                 arrow_t_pos=0.8, ):
        """

        :param from_port:
        :param to_port:
        :param editor:
        :param width:
        :param api_object:
        :param arrow_size:
        :param draw_labels:
        """
        GenericDiagramWidget.__init__(self, parent=None, api_object=api_object, editor=editor, draw_labels=draw_labels)
        QGraphicsLineItem.__init__(self)

        if isinstance(api_object, Transformer2W):
            if isinstance(api_object, Winding):  # Winding is a sublass of Transformer
                self.symbol = None
            else:
                self.symbol = TransformerSymbol(parent=self, pen_width=width, h=80, w=80)
        elif isinstance(api_object, VSC):
            self.symbol = VscSymbol(parent=self, pen_width=width, h=48, w=48)
        elif isinstance(api_object, UPFC):
            self.symbol = UpfcSymbol(parent=self, pen_width=width, h=48, w=48)
        elif isinstance(api_object, HvdcLine):
            self.symbol = HvdcSymbol(parent=self, pen_width=width, h=30, w=30)
        elif isinstance(api_object, SeriesReactance):
            self.symbol = SeriesReactanceSymbol(parent=self, pen_width=width, h=30, w=30)
        elif isinstance(api_object, Switch):
            if api_object.graphic_type == SwitchGraphicType.CircuitBreaker:
                self.symbol = SwitchSymbol(parent=self, pen_width=width, h=30, w=30)
            elif api_object.graphic_type == SwitchGraphicType.Disconnector:
                self.symbol = DisconnectorSymbol(parent=self, pen_width=width, h=30, w=30)
            else:
                self.symbol = DisconnectorSymbol(parent=self, pen_width=width, h=30, w=30)
        else:
            self.symbol = None

        self.scale = 1.0
        self.pen_style = Qt.PenStyle.SolidLine
        self.pen_color = QColor(0, 0, 0, 255)  # Black
        self.pen_width = width
        self.width = width

        self.setFlag(self.GraphicsItemFlag.ItemIsSelectable, True)
        self.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))

        self.pos1: QPointF = QPointF(0.0, 0.0)
        self.pos2: QPointF = QPointF(0.0, 0.0)

        self._from_port: Union[BarTerminalItem, RoundTerminalItem, None] = None
        self._to_port: Union[BarTerminalItem, RoundTerminalItem, None] = None

        # arrows
        self.view_arrows = True
        self.arrow_p_from = ArrowHead(parent=self, arrow_size=arrow_size, position=arrow_f_pos, under=False)
        self.arrow_q_from = ArrowHead(parent=self, arrow_size=arrow_size, position=arrow_f_pos, under=True)
        self.arrow_p_to = ArrowHead(parent=self, arrow_size=arrow_size, position=arrow_t_pos, under=False)
        self.arrow_q_to = ArrowHead(parent=self, arrow_size=arrow_size, position=arrow_t_pos, under=True)

        if from_port and to_port:
            self.redraw()

        if from_port:
            self.set_from_port(from_port)

        if to_port:
            self.set_to_port(to_port)

        self.set_colour(self.color, self.width, self.style)

    @property
    def parent(self) -> NODE_GRAPHIC:
        return self._parent

    @property
    def editor(self) -> SchematicWidget:
        return self._editor

    @property
    def api_object(self) -> BRANCH_TYPES:
        return self._api_object

    def delete_from_associations(self):
        """
        Delete this object from other associations, i.e. for a line, delete from the terminal connections
        :return:
        """
        self._from_port.delete_hosting_connection(graphic_obj=self)
        self._to_port.delete_hosting_connection(graphic_obj=self)

    def get_terminal_from(self) -> Union[None, BarTerminalItem, RoundTerminalItem]:
        """
        Get the terminal from
        :return: TerminalItem 
        """
        return self._from_port

    def get_terminal_to(self) -> Union[None, BarTerminalItem, RoundTerminalItem]:
        """
        Get the terminal to
        :return: TerminalItem
        """
        return self._to_port

    def get_terminal_from_parent(self) -> Union[None, BusGraphicItem, Transformer3WGraphicItem,
                                                FluidNodeGraphicItem, VscGraphicItem3Term]:
        """
        Get the terminal from parent object
        :return: TerminalItem 
        """
        return self._from_port.get_parent()

    def get_terminal_to_parent(self) -> Union[None, BusGraphicItem, Transformer3WGraphicItem,
                                              FluidNodeGraphicItem, VscGraphicItem3Term]:
        """
        Get the terminal to parent object
        :return: TerminalItem
        """
        return self._to_port.get_parent()

    def update_ports(self):
        """

        :return:
        """
        self._from_port.update()
        self._to_port.update()

    def recolour_mode(self) -> None:
        """
        Change the colour according to the system theme
        """
        super().recolour_mode()

        self.set_colour(self.color, self.width, self.style)

    def set_colour(self, color: QColor, w, style: Qt.PenStyle):
        """
        Set color and style
        :param color: QColor instance
        :param w: width
        :param style: PenStyle instance
        :return:
        """

        pen = QPen(color, w, style, Qt.PenCapStyle.RoundCap, Qt.PenJoinStyle.RoundJoin)

        self.setPen(pen)
        self.arrow_p_from.set_colour(color)
        self.arrow_q_from.set_colour(color)
        self.arrow_p_to.set_colour(color)
        self.arrow_q_to.set_colour(color)

        if self.symbol is not None:
            self.symbol.set_colour(color, w, style)

    def setToolTipText(self, toolTip: str):
        """
        Set branch tool tip text
        Args:
            toolTip: text
        """
        self.setToolTip(toolTip)

        if self.symbol is not None:
            self.symbol.setToolTipText(toolTip=toolTip)

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent):
        """
        mouse press: display the editor
        :param event:
        :return:
        """
        if self.api_object is not None:
            self.editor.set_editor_model(api_object=self.api_object)

    def remove_widget(self):
        """
        Remove this object in the diagram
        @return:
        """
        self.editor._remove_from_scene(self)

    def delete(self, ask=True):
        """
        Remove this object in the diagram and the API
        @return:
        """
        deleted, delete_from_db_final = self.editor.delete_with_dialogue(selected=[self], delete_from_db=False)

    def enable_disable_toggle(self):
        """

        @return:
        """
        if self.api_object is not None:
            if self.api_object.active:
                self.set_enable(False)
            else:
                self.set_enable(True)

            if self._editor.circuit.get_time_number() > 0:
                ok = yes_no_question('Do you want to update the time series active status accordingly?',
                                     'Update time series active status')

                if ok:
                    # change the bus state (time series)
                    self._editor.set_active_status_to_profile(self.api_object, override_question=True)

    def set_enable(self, val=True):
        """
        Set the enable value, graphically and in the API
        @param val:
        @return:
        """
        self.api_object.active = val
        if self.api_object is not None:
            if self.api_object.active:
                self.style = ACTIVE['style']
                self.color = ACTIVE['color']
            else:
                self.style = DEACTIVATED['style']
                self.color = DEACTIVATED['color']
        else:
            self.style = OTHER['style']
            self.color = OTHER['color']

        if self.symbol:
            self.symbol.setBrush(self.color)
            if self.api_object.active:
                self.symbol.setPen(QPen(ACTIVE['color']))
            else:
                self.symbol.setPen(QPen(DEACTIVATED['color']))

        # Set pen for everyone
        self.set_pen(QPen(self.color, self.width, self.style))

        self.arrow_p_from.setVisible(val)
        self.arrow_q_from.setVisible(val)
        self.arrow_p_to.setVisible(val)
        self.arrow_q_to.setVisible(val)

    def plot_profiles(self):
        """
        Plot the time series profiles
        @return:
        """
        # get the index of this object
        i = self._editor.circuit.get_branches().index(self.api_object)
        self._editor.plot_branch(i, self.api_object)

    def set_from_port(self, from_port: BarTerminalItem):
        """
        Set the From terminal in a connection
        @param from_port: TerminalItem
        """

        if self._from_port:
            # there was a port before, unregister it
            self._from_port.delete_hosting_connection(graphic_obj=self)

        # register the new port
        self._from_port = from_port
        self._from_port.add_hosting_connection(graphic_obj=self, callback=self.setBeginPos)
        self._from_port.update()
        self._from_port.get_parent().setZValue(0)

    def set_to_port(self, to_port: BarTerminalItem):
        """
        Set the To terminal in a connection
        @param to_port: TerminalItem
        """
        if self._to_port:
            # there was a port before, unregister it
            self._to_port.delete_hosting_connection(graphic_obj=self)

        # register the new port
        self._to_port = to_port
        self._to_port.add_hosting_connection(graphic_obj=self, callback=self.setEndPos)
        self._to_port.update()
        self._to_port.get_parent().setZValue(0)

    def unregister_port_from(self) -> None:
        """

        :return:
        """
        if self._from_port:
            self._from_port.delete_hosting_connection(graphic_obj=self)

    def unregister_port_to(self) -> None:
        """

        :return:
        """
        if self._to_port:
            self._to_port.delete_hosting_connection(graphic_obj=self)

    def setEndPos(self, endpos: QPointF):
        """
        Set the starting position
        @param endpos:
        @return:
        """
        self.pos2 = endpos
        self.redraw()

    def setBeginPos(self, pos1: QPointF):
        """
        Set the starting position
        @param pos1:
        @return:
        """
        self.pos1 = pos1
        self.redraw()

    def redraw(self) -> None:
        """
        Redraw the line with the given positions
        @return:
        """
        if self.pos1 is not None and self.pos2 is not None:

            # Set position
            self.setLine(QLineF(self.pos1, self.pos2))

            # set Z-Order (to the back)
            self.setZValue(-1)

            if self.api_object is not None:

                # arrows
                if self.view_arrows:
                    self.arrow_p_from.redraw()
                    self.arrow_q_from.redraw()
                    self.arrow_p_to.redraw()
                    self.arrow_q_to.redraw()

                if self.symbol is not None:
                    self.symbol.redraw()

    def set_pen(self, pen, scale: float = 1.0):
        """
        Set pen to all objects
        :param pen:
        :param scale:
        :return:
        """

        self.pen_style = pen.style()
        self.pen_color = pen.color()
        self.pen_width = pen.width()
        self.scale = scale

        pen.setWidth(self.pen_width / scale)

        self.setPen(pen)

        if self.symbol:
            self.symbol.set_pen(pen)

    def assign_rate_to_profile(self):
        """
        Assign the snapshot rate to the profile
        """
        self._editor.set_rate_to_profile(self.api_object)

    def assign_status_to_profile(self):
        """
        Assign the snapshot rate to the profile
        """
        self._editor.set_active_status_to_profile(self.api_object)

    def set_arrows_with_power(self, Sf: complex | None, St: complex | None) -> None:
        """
        Set the arrow directions
        :param Sf: Complex power from
        :param St: Complex power to
        """

        if Sf is not None:
            if St is None:
                St = -Sf

            Pf = Sf.real
            Qf = Sf.imag
            Pt = St.real
            Qt_ = St.imag
            self.arrow_p_from.set_value(Pf, True, Pf < 0, name="Pf", units="MW", draw_label=self.draw_labels)
            self.arrow_q_from.set_value(Qf, True, Qf < 0, name="Qf", units="MVAr", draw_label=self.draw_labels)
            self.arrow_p_to.set_value(Pt, True, Pt > 0, name="Pt", units="MW", draw_label=self.draw_labels)
            self.arrow_q_to.set_value(Qt_, True, Qt_ > 0, name="Qt", units="MVAr", draw_label=self.draw_labels)
        else:
            if St is None:
                self.arrow_p_from.setVisible(False)
                self.arrow_q_from.setVisible(False)
                self.arrow_p_to.setVisible(False)
                self.arrow_q_to.setVisible(False)

    def set_arrows_with_hvdc_power(self, Pf: float, Pt: float) -> None:
        """
        Set the arrow directions
        :param Pf: Complex power from
        :param Pt: Complex power to
        """
        self.arrow_p_from.set_value(Pf, True, Pf < 0, name="Pf", units="MW", draw_label=self.draw_labels)
        self.arrow_q_from.set_value(Pf, True, Pf < 0, name="Pf", units="MW", draw_label=self.draw_labels)
        self.arrow_p_to.set_value(Pt, True, Pt > 0, name="Pt", units="MW", draw_label=self.draw_labels)
        self.arrow_q_to.set_value(Pt, True, Pt > 0, name="Pt", units="MW", draw_label=self.draw_labels)

    def set_arrows_with_fluid_flow(self, flow: float) -> None:
        """
        Set the arrow directions
        :param flow: branch_flow (m3/s)
        """
        self.arrow_p_from.set_value(flow, True, flow < 0, name="flow", units="m3/s", draw_label=self.draw_labels)
        self.arrow_q_from.set_value(flow, True, flow < 0, name="flow", units="m3/s", draw_label=False)
        self.arrow_p_to.setVisible(False)
        self.arrow_q_to.setVisible(False)

    def change_bus(self):
        """
        change the from or to bus of the nbranch with another selected bus
        """
        self._editor.change_bus(line_graphics=self)

    def get_from_graphic_object(self):
        """

        :return:
        """
        if self._from_port:
            return self.get_terminal_from_parent()
        else:
            return None

    def get_to_graphic_object(self):
        """

        :return:
        """
        if self._to_port:
            return self.get_terminal_to_parent()
        else:
            return None

    def is_from_port_a_bus(self) -> bool:
        """

        :return:
        """
        if self._from_port:
            return isinstance(self.get_terminal_from_parent(), BusGraphicItem)
        else:
            return False

    def is_to_port_a_bus(self) -> bool:
        """

        :return:
        """
        if self._to_port:
            return isinstance(self.get_terminal_to_parent(), BusGraphicItem)
        else:
            return False

    def is_from_port_a_tr3(self) -> bool:
        """

        :return:
        """
        if self._from_port:
            if 'Transformer3WGraphicItem' not in sys.modules:
                # keep this here, we need an actual instance
                from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.transformer3w_graphics import \
                    Transformer3WGraphicItem
            return isinstance(self.get_terminal_from_parent(), Transformer3WGraphicItem)
        else:
            return False

    def is_to_port_a_tr3(self) -> bool:
        """

        :return:
        """
        if self._to_port:
            if 'Transformer3WGraphicItem' not in sys.modules:
                # keep this here, we need an actual instance
                from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.transformer3w_graphics import \
                    Transformer3WGraphicItem
            return isinstance(self.get_terminal_to_parent(), Transformer3WGraphicItem)
        else:
            return False

    def is_from_port_a_fluid_node(self) -> bool:
        """

        :return:
        """
        if self._from_port:
            return isinstance(self.get_terminal_from_parent(), FluidNodeGraphicItem)
        else:
            return False

    def is_to_port_a_fluid_node(self) -> bool:
        """

        :return:
        """
        if self._to_port:
            return isinstance(self.get_terminal_to_parent(), FluidNodeGraphicItem)
        else:
            return False

    def get_bus_from(self) -> Bus:
        """

        :return:
        """
        return self.get_from_graphic_object().api_object

    def get_bus_to(self) -> Bus:
        """

        :return:
        """
        return self.get_to_graphic_object().api_object

    def get_cn_from(self) -> Bus:
        """

        :return:
        """
        return self.get_from_graphic_object().api_object

    def get_cn_to(self) -> Bus:
        """

        :return:
        """
        return self.get_to_graphic_object().api_object

    def get_busbar_from(self) -> BusBar:
        """

        :return:
        """
        return self.get_from_graphic_object()._api_object

    def get_busbar_to(self) -> BusBar:
        """

        :return:
        """
        return self.get_to_graphic_object()._api_object

    def get_fluid_node_from(self) -> FluidNode:
        """

        :return:
        """
        return self.get_from_graphic_object()._api_object

    def get_fluid_node_to(self) -> FluidNode:
        """

        :return:
        """
        return self.get_to_graphic_object()._api_object

    def get_fluid_node_graphics_from(self) -> FluidNodeGraphicItem:
        """

        :return:
        """
        return self.get_from_graphic_object()

    def get_fluid_node_graphics_to(self) -> FluidNodeGraphicItem:
        """

        :return:
        """
        return self.get_to_graphic_object()

    def connected_between_buses(self):
        """

        :return:
        """
        return self.is_from_port_a_bus() and self.is_to_port_a_bus()

    def connected_between_bus_and_tr3(self):
        """

        :return:
        """
        return self.is_from_port_a_bus() and self.is_to_port_a_tr3()

    def conneted_between_tr3_and_bus(self):
        """

        :return:
        """
        return self.is_from_port_a_tr3() and self.is_to_port_a_bus()

    def connected_between_fluid_nodes(self):
        """

        :return:
        """
        return self.is_from_port_a_fluid_node() and self.is_to_port_a_fluid_node()

    def connected_between_fluid_node_and_bus(self):
        """

        :return:
        """
        return self.is_from_port_a_fluid_node() and self.is_to_port_a_bus()

    def connected_between_bus_and_fluid_node(self):
        """

        :return:
        """
        return self.is_from_port_a_bus() and self.is_to_port_a_fluid_node()

    def should_be_a_converter(self) -> bool:
        """

        :return:
        """
        return self.get_bus_from().is_dc != self.get_bus_to().is_dc

    def should_be_a_dc_line(self) -> bool:
        """

        :return:
        """
        return self.get_bus_from().is_dc and self.get_bus_to().is_dc

    def should_be_a_transformer(self, branch_connection_voltage_tolerance: float = 0.1) -> bool:
        """

        :param branch_connection_voltage_tolerance:
        :return:
        """
        bus_from = self.get_bus_from()
        bus_to = self.get_bus_to()

        V1 = min(bus_to.Vnom, bus_from.Vnom)
        V2 = max(bus_to.Vnom, bus_from.Vnom)
        if V2 > 0:
            per = V1 / V2
            return per < (1.0 - branch_connection_voltage_tolerance)
        else:
            return V1 != V2



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations
from typing import TYPE_CHECKING, List
from PySide6.QtCore import Qt, QPointF
from PySide6.QtGui import QPen, QCursor
from PySide6.QtWidgets import (QGraphicsLineItem, QGraphicsItemGroup, QMenu,
                               QGraphicsSceneContextMenuEvent)
from VeraGrid.Gui.messages import yes_no_question, error_msg, warning_msg
from VeraGrid.Gui.gui_functions import add_menu_entry
from VeraGrid.Gui.Diagrams.generic_graphics import (GenericDiagramWidget, ACTIVE, DEACTIVATED, OTHER, Square, Circle,
                                                    Polygon, Condenser)

from VeraGridEngine.Devices.types import INJECTION_DEVICE_TYPES

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget
    from VeraGrid.Gui.Diagrams.SchematicWidget.Substation.bus_graphics import BusGraphicItem
    from VeraGrid.Gui.Diagrams.SchematicWidget.Fluid.fluid_node_graphics import FluidNodeGraphicItem

    NODE_GRAPHIC = BusGraphicItem | FluidNodeGraphicItem


class InjectionTemplateGraphicItem(GenericDiagramWidget, QGraphicsItemGroup):
    """
    InjectionTemplateGraphicItem
    """

    def __init__(self,
                 parent: NODE_GRAPHIC,
                 api_obj: INJECTION_DEVICE_TYPES,
                 device_type_name: str,
                 w: int,
                 h: int,
                 editor: SchematicWidget):
        """

        :param parent:
        :param api_obj:
        :param device_type_name:
        :param w:
        :param h:
        """
        GenericDiagramWidget.__init__(self, parent=parent, api_object=api_obj, editor=editor, draw_labels=True)
        QGraphicsItemGroup.__init__(self, parent)

        self.w = w
        self.h = h
        self.glyph: Square | Circle | Polygon | Condenser | None = None
        self.scale = 1.0
        self.device_type_name = device_type_name

        # Properties of the container:
        self.setFlags(self.GraphicsItemFlag.ItemIsSelectable | self.GraphicsItemFlag.ItemIsMovable)
        self.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))

        self.width = 4

        # line to tie this object with the original bus (the parent)
        self.nexus = QGraphicsLineItem()
        self.nexus.setPen(QPen(self.color, self.width, self.style))


    def set_glyph(self, glyph: Square | Circle | Polygon | Condenser):
        """

        :param glyph:
        :return:
        """
        pen = QPen(self.color, self.width, self.style)
        self.glyph = glyph
        self.glyph.setPen(pen)
        self.addToGroup(self.glyph)

        self.setPos(self.parent.x(), self.parent.y() + 100)
        self.update_nexus(self.pos())

        self._editor.add_to_scene(self.nexus)

    @property
    def parent(self) -> NODE_GRAPHIC:
        return self._parent

    @property
    def api_object(self) -> INJECTION_DEVICE_TYPES:
        return self._api_object

    @property
    def editor(self) -> SchematicWidget:
        return self._editor

    def delete_from_associations(self):
        """
        Delete this object from the bus or other parent hosting it
        """
        self.parent.delete_child(self)

    def get_associated_widgets(self) -> List[GenericDiagramWidget | QGraphicsLineItem]:
        """
        Get a list of all associated graphics
        :return:
        """
        return list()

    def get_extra_graphics(self):
        """
        Get a list of all QGraphicsItem that are not GenericDiagramWidget elements associated with this widget.
        :return:
        """
        return [self.nexus, self.glyph]

    def recolour_mode(self):
        """
        Change the colour according to the system theme
        """
        super().recolour_mode()

        pen = QPen(self.color, self.width, self.style)
        self.nexus.setPen(pen)
        self.glyph.setPen(pen)

        return pen

    def enable_disable_toggle(self):
        """

        @return:
        """
        if self.api_object is not None:
            if self.api_object.active:
                self.set_enable(False)
            else:
                self.set_enable(True)

            if self.editor.circuit.has_time_series:
                ok = yes_no_question('Do you want to update the time series active status accordingly?',
                                     'Update time series active status')

                if ok:
                    # change the bus state (time series)
                    self.editor.set_active_status_to_profile(self.api_object, override_question=True)

    def set_enable(self, val=True):
        """
        Set the enable value, graphically and in the API
        @param val:
        @return:
        """
        self.api_object.active = val
        if self.api_object is not None:
            if self.api_object.active:
                self.style = ACTIVE['style']
                self.color = ACTIVE['color']
            else:
                self.style = DEACTIVATED['style']
                self.color = DEACTIVATED['color']
        else:
            self.style = OTHER['style']
            self.color = OTHER['color']

        self.glyph.setPen(QPen(self.color, self.width, self.style))

    def update_nexus(self, pos: QPointF):
        """
        Update the nexus line that joins the parent and this object
        :param pos: position of this object
        """
        parent_pt = self.parent.get_nexus_point()
        self.nexus.setLine(
            pos.x() + self.w / 2,
            pos.y(),
            parent_pt.x(),
            parent_pt.y(),
        )
        self.setZValue(-1)
        self.nexus.setZValue(-1)

    def delete(self):
        """
        Remove this element
        @return:
        """

        deleted, delete_from_db_final = self.editor.delete_with_dialogue(selected=[self], delete_from_db=False)


    def plot(self):
        """
        Plot API objects profiles
        """
        # time series object from the last simulation
        ts = self.editor.circuit.time_profile

        # plot the profiles
        self.api_object.plot_profiles(time=ts)

    def mousePressEvent(self, QGraphicsSceneMouseEvent):
        """
        mouse press: display the editor
        :param QGraphicsSceneMouseEvent:
        :return:
        """
        self._editor.set_editor_model(api_object=self.api_object)

    def change_bus(self):
        """
        Change the generator bus
        """
        idx_bus_list = self._editor.get_selected_buses()

        if len(idx_bus_list) == 2:

            # detect the bus and its combinations
            if idx_bus_list[0][1] == self.api_object.bus:
                idx, old_bus, old_bus_graphic_item = idx_bus_list[0]
                idx, new_bus, new_bus_graphic_item = idx_bus_list[1]
            elif idx_bus_list[1][1] == self.api_object.bus:
                idx, new_bus, new_bus_graphic_item = idx_bus_list[0]
                idx, old_bus, old_bus_graphic_item = idx_bus_list[1]
            else:
                error_msg("The bus to change has not been selected!", 'Change bus')
                return

            ok = yes_no_question(
                text=f"Are you sure that you want to relocate the bus from {old_bus.name} to {new_bus.name}?",
                title='Change bus')

            if ok:
                self.api_object.bus = new_bus
                new_bus_graphic_item.add_object(api_obj=self.api_object)
                new_bus_graphic_item.update()
                self._editor.remove_element(device=self.api_object, graphic_object=self)

        else:
            warning_msg("you have to select the origin and destination buses!",
                        title='Change bus')

    def rescale(self, scale: float = 1.0):
        """

        :param scale:
        :return:
        """
        self.scale = scale

    def get_base_context_menu(self) -> QMenu:

        menu = QMenu()

        add_menu_entry(menu=menu,
                       text="Active",
                       function_ptr=self.enable_disable_toggle,
                       checkeable=True,
                       checked_value=self.api_object.active)

        add_menu_entry(menu=menu,
                       text="Change bus",
                       function_ptr=self.change_bus,
                       icon_path=":/Icons/icons/move_bus.svg")

        add_menu_entry(menu=menu,
                       text="Plot profiles",
                       function_ptr=self.plot,
                       icon_path=":/Icons/icons/plot.svg")

        menu.addSeparator()

        add_menu_entry(menu=menu,
                       text="Delete",
                       function_ptr=self.delete,
                       icon_path=":/Icons/icons/delete_schematic.svg")

        return menu

    def contextMenuEvent(self, event: QGraphicsSceneContextMenuEvent):
        """
        Display context menu
        @param event:
        @return:
        """
        menu = self.get_base_context_menu()

        menu.exec(event.screenPos())



# MIT License
#
# Copyright (c) 2021-2022 Yunosuke Ohsugi
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module setting up ui of dock window."""
from __future__ import annotations

from VeraGrid.ThirdParty.qdarktheme.qtpy.QtCore import Qt
from VeraGrid.ThirdParty.qdarktheme.qtpy.QtWidgets import QDockWidget, QMainWindow, QTextEdit, QVBoxLayout, QWidget


class DockUI:
    """The ui class of dock window."""

    def setup_ui(self, win: QWidget) -> None:
        """Set up ui."""
        # Widgets
        left_dock = QDockWidget("Left dock")
        right_dock = QDockWidget("Right dock")
        top_dock = QDockWidget("Top dock")
        bottom_dock = QDockWidget("Bottom dock")

        # Setup widgets
        left_dock.setWidget(QTextEdit("This is the left widget."))
        right_dock.setWidget(QTextEdit("This is the right widget."))
        top_dock.setWidget(QTextEdit("This is the top widget."))
        bottom_dock.setWidget(QTextEdit("This is the bottom widget."))
        for dock in (left_dock, right_dock, top_dock, bottom_dock):
            dock.setAllowedAreas(
                Qt.DockWidgetArea.LeftDockWidgetArea
                | Qt.DockWidgetArea.RightDockWidgetArea
                | Qt.DockWidgetArea.BottomDockWidgetArea
                | Qt.DockWidgetArea.TopDockWidgetArea
            )

        # Layout
        main_win = QMainWindow()
        main_win.setCentralWidget(QTextEdit("This is the central widget."))
        main_win.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, left_dock)
        main_win.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, right_dock)
        main_win.addDockWidget(Qt.DockWidgetArea.TopDockWidgetArea, top_dock)
        main_win.addDockWidget(Qt.DockWidgetArea.BottomDockWidgetArea, bottom_dock)

        layout = QVBoxLayout(win)
        layout.addWidget(main_win)
        layout.setContentsMargins(0, 0, 0, 0)



# MIT License
#
# Copyright (c) 2018 Ross Wilson
# Copyright (c) 2024, Santiago Peñate Vera
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Source of the 'error' and 'pending' tiles data.

The code here is generated from the:
    error_tile.png
    pending_tile.png
files in the pyslipqt/examples/graphics directory using:
    /usr/local/bin/img2py <file>

The captured output is cleaned up and saved in this file as data.
"""

import base64


# the 'pending' tile data
def getPendingImage():
    """Generate 'pending' image from embedded data."""

    return base64.b64decode(
        "iVBORw0KGgoAAAANSUhEUgAAAQAAAAEACAIAAADTED8xAAAACXBIWXMAAAsT"
        "AAALEwEAmpwYAAAAB3RJTUUH4gIJAxggK6MKCwAAIABJREFUeNrsvWlwXNeV"
        "JnjOufdtmUgsiY0ACJIAuK8StVGWbJVU8tJVtmZctqpcsstR1TUTMRUVPdM/"
        "JmY6JmI6OmJ+TfT8618z86NnpvYql2W7ynZJlmXtCzdREkESJEGCWAgQ+5bL"
        "e+/ec+bHSyQBrkmKbovg+yIlIm9mvnzv5fnuPefcs+DLL78MKVI8qKD0FqRI"
        "CZAiRUqAFClSAqRIkRIgRYqUAClSpARIkSIlQIoUKQFSpEgJkCJFSoAUKVIC"
        "pEiREiBFipQAKVKkBEiRIiVAihQpAVKkSAmQIkVKgBQpUgKkSJESIEWKlAAp"
        "UqQESJEiJUCKFCkBUqRICZAiRUqAFClSAqRIkRIgRYqUAClSpARIkSIlQIoU"
        "KQFSpEgJkCJFSoAUKVICpEiREiBFipQAKVKkBEiRIiVAihQpAVKkSAmQIkVK"
        "gBQpUgKkSJESIEWKlAApUqQESJEiJUCKFCkBUqRICZAiRUqAFClSAqRICZAi"
        "RUqAFClSAqRIkRIgRYqUAClSpARIkSIlQIoUKQFSpEgJkCJFSoAUKVICpEiR"
        "EiBFipQAKVKkBEiRIiVAihT3I3R6C2rB8vT08tJkuDSzVAhLYRhFRsSCgFak"
        "HTcIvFxDkMnk65ra/Fw+vV0pAdYDwuW5qaGBy9OzczNzU/NLc4sLS4tLYVSI"
        "CssmjsRYACBHa8f1MlkvyOXqcvlcfWtTrqU139re3rWpB/zG9DamBLjPYIyZ"
        "vnj8/IXLI6NXRq6Mz89cDpcKhESoEZGQiNDVfuXOCUDM4dxSeWZ+DuGijQEh"
        "aGxoynds6tywuauju6enbfNOolTV/JwCX3755fQuVFCeHzz98ZlzQ6cHL82M"
        "XzShUeQopSjnou86WU/5LvkkHoIG1CAAGCPEArHYguUoipcjLoVciERsbGIn"
        "G3R09+3Y0r1vZ0/vtp1lJ9WOUgJ8LhHH8djJIx+dOv3pmYGlqXFNmrSLdY6X"
        "r3MafW4D2yzShKY5Ludi9tn4pvpZVVK6qINFV80TzSBOA01JNF+M54u8FMU2"
        "RiVN7Zv379x64JGHO3v2ImJ6w1MCfI4wNXLuyAfvH/v09OLUZU0aPcfNZ/y2"
        "et4CZiOUu0tR1rpAzERAifiuFmJBAQYRYWBQ1ojoRZ0Z8vUI4jCXphbMfFmi"
        "0JK0dGw59PDuh598KtfUlUpeSoDfPKIoOnf87dfePXx56CxZQN9zm7Pepnqz"
        "TcIdYbkhdkKHiJQG6wIqJEcIEREBEZTEKCioGEQEGMQKRMAhihUr1mgTzLru"
        "GVefk+LIAs+VTFQm3+3dvverX3xs4+5DqWGQEuA3ieLs5bdf/9UHJ44vT0+7"
        "rk+NXmZL3uyHYl8onriiRJPyATwBDQJAgsAAAKwEDSAiKwEAEhSU1cuCWLEh"
        "qjJwxBFaXULvjKs+luLoLC9H1oa5lq5nn3ri8ae+5GSbUhH8zUJ95zvfeQAv"
        "e/byhX98+Z+OHf8gXi66uUywMa+fypafisM245IiT0EWtY/gCApiTGgQAAQE"
        "AMgQMYEAAqAhYCQmZAIDYBGZUEhpRA/JJW0VCpY6Yt5EgV9HBYIYivOzQ5dH"
        "lhcWNnW1O0EulcKUAP9Fcfnimb//wcsXzn6ihFRjkNnZar5Epa2Rg9pxNeZQ"
        "uQSMwiiMyhIKoiAwVASdAQFREC2iIDFVniYjjIAgIiCAhBQgatJGMUK8MXZa"
        "Am85I5ExS+Wx8bHJqfkt3Z1+tj4VxJQA/6Wkf/DE3/7Dj8YuDTjKU/mM/0Te"
        "PM62ybrgSB1KBkSQDAgAsSATMwgDCAIjMSIgMFgAYBRBEkRGISFDyChKQKDK"
        "EGAAC0iIHoImXdRhQ4wd5NmsXbZYMpNTY+NXZvu62/xcqgulBPj1Y/zCwN/9"
        "44/HL51znMBpybhfagj3MiChT5BB1KgNogAIgCAxCKAGJEAEEUHFaBGEEQGF"
        "BQRYQBiASQREQNkKQxAwUZNYEEWQUSmEACjUTAKd4DoZnjNQlpmZy+NTC9u2"
        "bvSCdB1ICfBr1ftHLvz9D18euzjguIHbllHP1Ee97BhH16FkgACEQTEioxUg"
        "BhZUDMAADMyoAIABDGpAFnCYFKMCBCEFAIyKgQWFERiFwQKDEAqKgGIEA4hI"
        "PhAglSluYTeb4akYS3Z2dmJurtDbu8X1M6lEpgT4taA0N/7yj/9p8Mwnjvac"
        "loz6Yr3ZJK7RVI82ABQEiwhIFoCBGJFRAIEAmIghmccJEAQRQASBkQHAIlok"
        "BrKICGCSDwoiakZiAEE0KICMK1qTD6KIihDn2Q8ydsZI0UzNTUgU9vVuRsdP"
        "hTIlwD1GGIZvvvLTYyc+VEyqKXC+0GA3g8uaciQeggVlVKKxCKJ1BVwRV8Bh"
        "BCCLYkgbIqOEQSGIIQAQFkQkEUFAEUSwFjSi1cweo8viiWgRABTEGNGiTgxr"
        "AXBBo8ISxE3iSYbnQi7E47PT9b7XuWXb3V3jm7/8p6nJ0ShcQFLWguu6qXDX"
        "ggciGO78iXc++Ogol43TEPgHmso97MREOQREHQIAMoqtN+yyOx7QCMqcQARA"
        "QDmENgq3FiPNzqKLMYpFUoKGRBCtAAMqAUAEkIa4pKRhMMfjIvMMFkEDNSB2"
        "4PL2RQCgeRcsakBrQQJRRtl5ifeyv9wcnpsuzS/88p0Puza2d2x95C6ukdlM"
        "Tc4sl2hoeD4ZqaurC4IgCIJczvP9esdxXNdVSqVC/2ARYHr0/K/ePVKcnXXq"
        "sv6W5nibaEPkOICCMTJA1Bj5BUe97ZVOh9PzV5aWClEhtLEgopPRmbpMY2Nj"
        "XW8dPgrhhrKe8RAFAEAQEQQRIzDNoSpq/5dB6VT58uzY0uJyWIiZWWlyAydb"
        "X9fY2Oj3+vg4hBvKPOe4VolmCCAou+UwNAfAWWiwocxODP3yzSPfbN2SbWi+"
        "08tEUH42BwCImOwxh2EYhuHi4uLk5NXwDa0pCDzXdR0pOFqhk/M8x3U0ub4D"
        "ShNYnVVKPTi71OucAHEcH37/vZGLA9rx3bY6+xCIEo0aHcSYxLHgiH/EWz66"
        "ODJ0eWls6fojzMH8GFwO+oPOUx0tj7eYh62ySBGBBraoCOL6yPsoWH5n+dL5"
        "c0vjNzjCNMxeguHMyUznxx2tD7XIk2KCWC85rAEDUJFjspE85OplHybN6TMf"
        "bz268dBvf/1Or9Sray6Xo+vHRQQRE1Yk/4ahLS9PxFFIAKCmCIEIBFAhasWg"
        "CC2SIiImIqWUQnR85Vgrkqwepm3rl1IC3B8YHTh8/OQptIgNHu0ITJ0Qk3IV"
        "GLLZWC2p+G179tiZxRXRr6+v37JlS76lJQgCY8zi/Pzw8PD4+HhprjR49MKV"
        "0cmd4zv0F7XJGBUqQrGO1a+75944N3NhNjlCU1NTT09PPp93PC8Ow7m5ueHh"
        "4ampqeJ08fz04OWhy9subM9+ORtuKTrTvjjieEoWlW217pZ6Lkbl+eUPP/q0"
        "d9v2tk3b7+hKibRIWEuoaRwu2LhEQIBAAqiQSAgQkISUAlEKCQEdhYIAAAI2"
        "ioEQOQYkim26AtwfMMX5Y8dPL0yOu5lM0NEQdwvGqB0NAOJYmqTCa4UzR8+a"
        "ogGAvr6+np6e7du3d3d3t6wQYGFhYWRk5Ny5c+fPn+/v71+eWP7k1U/2lvd6"
        "z7niCHuMb+On/3KyOF0EgK6url27du3evXvTpk35fN513TAM5+fnR0ZGLly4"
        "MDQ8/MmJE8WZ0qdvfbJ5YfOGf9Ue7wy9hAOOY8s23iHeZD0XzPjwuU+PHXt2"
        "49Zfhx4SRws2KpGiJK4pxXomwOCpj/oH+h3l6EbfbCUgISEiEhSYh8VXF899"
        "OGhjW1dXt3v37ueff/6hhx5yHGf1ERoaGjZt2vTUU0+dPHnyxz/+8ZkzZ5aX"
        "l0++dXK/vz96Mfbf9E794lQi/bt27fr617/+2GOPrT5CNpvN5/O9vb3PPPPM"
        "qVOn2lpaTpw4MT09ffH4kDWmi7qiLZG76CpHqVBFgdGbXJx1ZC76eGBw18VT"
        "nX177+0NsdGCicsKUVgQ0rSEde0GjaLovbfePH+6X2cDv7vJ7gSMUDkKNQJD"
        "+fXwzAdnbWRbWlqefPLJP/7jP+7r67uFh6StrW3fvn1j4+OzMzPlQrlULrVL"
        "65VjV6bOTwPAnj17/uzP/mzPnj23OEJra+v+/fuVUtPT04uLiwtXFoNyUNeV"
        "E1+QlbBI0XIjOHOaF6Olxel8Q+OW7Xtqv96xsbE4jnEFVau3agCYeNGEJSQk"
        "RMCr/yEhISAgrPxNyUtq5XVAJCFEFABEZK5r7UkJ8HnH2OkPX3/3iCmVdEuA"
        "e3wOhICUUuijPcpn3huIl+PGxsann376u9/9bi53+5DMIAh27dw5ODg4MTER"
        "LUfzo/PL08umbPbt2/fd7363t7f39qut1jt37lRKXb58eWlpaWFyoVW10HYi"
        "gyDAsVgtuqzNfMSFstFB7+b2TF3+nhDAhIs2DpEIEVICrLGd1uvSdv7C8MLU"
        "CDqO25S1rYIxIiIQmEkePTlani0DwP79+1988cUgCGo8Zj6ff/755zdu3MiG"
        "l68USvNlrfX+/ft37txZ+4k9//zzTz31VENDg43tuf5BGRB2GRWSJgwl3she"
        "U6Adf3T4/PC5s5/9PiBiFC3HYSFVeR4gAsxdGbkwOgkWMOtIXif2Hgqih+Z4"
        "NH1+GgAOHjz4wgsv1NXV3dGRH3nkkb5t21Y/feihh+5szVXq61//esKZxdHF"
        "xeOLGCMKEhEycQao3kdflRYWLo2Ox3Fcq35vb6x9mbhoowJgmn32IBHgyvC5"
        "scuXlHLcep83AJYBgYDAztmpiWkbsdZ69+7dfX19d3pkz/P6enq0rjgPmltb"
        "t27deqcHqa+vf/rppzdu3AgAE+NXzDkrKISICjEUbgNV53quNzQyMTN8usZj"
        "BvoXgs51Pp9SVF5M5/4HjgATV6aXpyfRc6jO5aygAWQADfa0XbiykEz/+/fv"
        "v7uDNzQ0tLS0JHN5S/4uK508+uijyUqyOLYUno+AIEmpkQhsk6g6l0jPzoxO"
        "jF+u1ehfPOzKB7zKrReXCyZeFEnF/wEjgFmcuDwxCwLga8gptgIAIgIA5dli"
        "caoEAC1tbdu23WXYWV1jY319PQBks9na7Ydr4Pv+1t5e13VtbAuLixIKAJBC"
        "tMiOYOCAr0sLS5NXZu/A8bX8nqc+FHESzcfEi7+W+7u+POfrcB9gZnZ8cm5W"
        "KUdnNNcDxklmFnIoy4VSInwd7e23Psj8/Pzhw4eXlpZ6tm3bt2vXau++S5S4"
        "O5VSVV2oisHBwZMnTyqlDhw40N3dfYuvaG1t3bBhw/Dw8HKhkJ9tpnpKJiQ2"
        "AFmFnoNLODE7X1y8nKnvrHE2ixfe9RqgFB4w0RLpVO9/IAkwPTE7Nz+lSJGv"
        "bU4gqWGlAJckLkeJM6ep6VYpiOPj43//gx+c/OST2dnZbdu2zX35y88++2x1"
        "XzYJMgOAcrlcLpdXf/D4J5+88tOfnjhxgoguXrz4la98ZceOHTf7lpaWlpaW"
        "luHh4bAYyYxIHUCMAAAxcoMon4j0zNzc3MSVWgmQaD5L78YcCe5KN3sfUALM"
        "Ly6XF2a18snVxhMqAipEES6jMQYAMplMJnOr3Ks333zznbfeSt48MDAQBEF3"
        "d/f27ZXgnIWFhYWFBQAolUoJExIUCoVXf/azw4cPJ09ff/11EdmyZYvneTf8"
        "luppmMhISUQEE1cVM/uIviaixcXZxeX5Oy2jFdCREEDMbtQ2FfEHzgYoLBVM"
        "ZMAn0IQMYjFJaxS+OiMmJsGNP14oAEAi/QmGhoYuXbpUfTo1NTUzM1NxN01N"
        "Vd85ODiYfLaKUqk0MjJS00kziBVkEEEwII6gJlSqtDi3MFe4i5vgwRGSU2LT"
        "6P8HjAAYLhTKJQRErcBBiBFFJBEvDYkaUywWi8XiTafPIJibm1s90tjYWN0u"
        "mJ2dTab/ChmuXDlz5kzVO3TNroKfyTQ0NNzsi6qnQZrQQbAiVlAEAAQFHYWu"
        "CoulsFy4u1vhuUcU9v86FnlmTgnwOcXSUlwox0gKCFkDJC4gK2IAXNCeSoR4"
        "fn7+pneE6Kmnnjp48GDyNJfL7d27d9++fcnTkydPVpUcAPj0009Pn6646ru7"
        "u/fs2dO+Yl5v3rz5wL59ra2tN/uiqamp6elpAPACR+oZDXCyUolAjEk9dgAo"
        "l81d3w1fH1VyjM095YC51fqZ2gC/YRgul+MIEVEhKAQjYgURgYED8QMvMV5H"
        "R0ettTeLXdu/f7/WetOWLXEcd7S3P/7448nUvry8/MHhwzMzM6TJz7umzNFi"
        "dPbs2cHBwb6+PiJ65plncrnc4OAgIu7Zs+fWm8Sjo6Ojo6MA4Gd8zCHHjFbE"
        "ClgUEiAQQkIK49AYc727qaapWiBwPy4bFHkM4N7YA5a91Aj+/MLGoTVWEQIi"
        "oIAVZAIUAcYyZuozQatXmgpHR0f7+/tvsRe2e/fu3bt3XzP4i1/84sypUwCQ"
        "afN792+5PDQxvTh77Nixxnz+X//xHwdBkMvlnnnmmWeeeea25zk5OTkxMcHM"
        "ylPZXJZF0FSK7AIDMQgBIBBRbITiZdB32WyGGXx9IrLA9nH1mW3isJDzGv/f"
        "V078+9955LVUBfo8gplZGACQoKJNJzYAg5RBdTjZ1iwA9Pf3v/Pee1EU1X7k"
        "995774033pibm0OFrZubcZdq62xxc46InDh+/Gc/+5m1dyBe77///rvvvgsA"
        "ua6s35mBMrABsFipMs1X1QxrrOXPqnL4dMLRR6x1PstBSsVc44a/JBhKjeDP"
        "8QrAwpYBQBgABQGQKw82DI7kNzQ6dRoAPv7oo1deeaV2ef3hD384PDwMAPlt"
        "DfVbGnEZ3B6vbUcrEExPT7/++us/+clPVntFb4Fjx46988475XIZFbZsyFML"
        "SARgRThxAyXFGKXqsPrM8g+M4NFHvvrAyl0qMFEx07Thb0rRIADAlVQF+twS"
        "epV/InH+iMBK/XLgkngdQfP2/MSJycnJyddee42Zv/rVr/r+TctRLS4uvvrq"
        "q++8807iCa3rynZt74SsSAjgQr6nuVQsz5yZvXz58s9//vPJycmvfe1rmzdv"
        "vsUC9d577/3zP//z+fPnASC/tbF+cwOHAixIABbEALIAreT2IsC9Mzg1nSCA"
        "mJ8kdQeGNSkOTTbf+vfl6HxllqlXKQE+p0Ba0UNQiMEyEACIACIyAIAoaN2S"
        "D5fLc2cXh4eHf/rTnw4PDx86dGj79u3XuCyHhoYuXLhw5MiRU6dOLS4uAkB2"
        "Y6Zn30ZoISkCoGARIScbd3QA88zZ+ampqVdffXV+fn7Hjh07d+7ctm3baiPb"
        "GHPq1KnjJ04cO3IksX0zGzJd27okJ1IGJBBGsYCCLFhpOfNrcDZqOYEOxvZJ"
        "wFoVtnKcaWz4hzA+B+sR63AnuJrtKsDClOgSyJAsAxAD1FHn9g0AMD+4NDU1"
        "9frrr589e/bhhx/WrpsNAiKy1ibbvZ988kni9UeFjVvru/raoVVDQYBAWAAQ"
        "CoIN2Lm70/XdqbOzpmg++OCDY8eOHThwoKWlpbn5anmfyenp8bGxU6dOJU70"
        "uq5s795N1I5QQkARTs4X2AqCMFeacdQ+/ZdLVGPsg4MfkYLIfoHw9utAbIOm"
        "hn+MymfJvSoyNsKUAJ9X6VdOEgTGVpSt7IIRojAAJuIkUBZqUht3d/p1s7MX"
        "5sP5qOqRRESl1OptYCDItAfNGxsbuxswUFBmEUw8igKMiFxkDLB1R2uQ86aG"
        "ZwtjpTiOjx49Wvk0EazdOdIZ3bSlvqN3A7UqKYIAAwAhCoHYFQOABFCS4CPU"
        "SpnbC9zpiwd6O0843u0ZwwgKPvJdKMdPu3ArDoQ225j5QWwH1nHthPV2Za7n"
        "u45rWZQIxCKcVDkXAECpZoaglAECbN3eUlefmZ1aKM2VwrnYlqyIJNKPCt16"
        "7Ta62cZMU3uDbnKAQSLGin29YmYkibchCEFddy7bmJnfsLg8VyjPhdFCbCOu"
        "ij655DW6mbzf2FJf11kPPkhBAEAIAMCKoCBYARZARkEwwJZZ2HOUZLO3v3Kh"
        "0xf37OnrJ6+mVUPJR4GHUfy0vgkH4lJQn/uByACsa6w3AmSculzgCVuOGWKL"
        "oCv+RERhhtV1o2IQkKAj6GoPovmotFSKwtjGNpm2tau9rJdp8CBQyJDE619V"
        "SQhABAQBOXHbAIMYQJ/yfU2N5YZoPioVSnFkrGEAUJoczwmygZd3wSWIQUoV"
        "7iBXirdVDiICAqCAjSELzJz1vGuKtdzCXu0f3LOnr//264ACtqDlOLlQNl9S"
        "dK3zKgzrsrkfWh4g5waOQuunKtDn9noaG5vqsyxMBiQyQmsV3Wv28AUlFEBx"
        "6l0374KquItgpfBtMutfn1SFViApEXrNS4bFABJ4zZ7X5qOGajoyiEgsACBl"
        "Fq5sU4ggEiAAWyFEQREGAmArEjFE7Pp+XUOtWcsioohOX9y3s++k69zeHmAG"
        "guMZDyLzJaKrycdRmMnWvUx05qafLEBKgM8vWlryQUODRJZLlpJwfUQQEBC8"
        "Rl4RWARBxDCYVfRYaROTzPW8Mu9XP5zkmOF1UaWVJwYBrURYKaNbeT9WVCYU"
        "EQSp6PpiK6xjRgJAEUbAInPZMJuG1o62ljvIumQBBDwzuHfPtk/JqUkXInvc"
        "dyA0z2g0ABCbIOP9SOkzt1pmTXHdSMs6DIfe0LWpfcNmtpEtxLDITCDMyfaq"
        "FUkeImJFmBmlqnjI6qkURFCSbeTKH5i8Z9UjSba8kSUuKyRaMz0DijCzFQC2"
        "VkBErIDY5Nww2cUWESVQsFyycRx3dHS1dt1xER5Eder8/rikqAZVhRmIjwfu"
        "m4ZdNq6n/0mrM7fxwK6jxpbrcAVo3bRj66Zjw+dOupFwIZR6DyBxNTICJVK5"
        "YpmCMAgwrtlAS7zyFf/M2gUj8cvI1YmfCarrAMpVjSgZoVVfhii2SoWkwXyi"
        "BQkSiVgBQgJgwFB4KcZQSFNfV2uu5W7ayiOqgYv7dvR9GmRsTRwwx7M+CZcc"
        "PM18y3nRgOuU0xXg8wul1M5dvfVtGyWK7XxMC1ZQwAoICjOseohhkVuNw0qP"
        "sGpKTRJZtHpeT+IXIIljq64PiRQyIK8yP1ZeQk5ItPLOJHyJmQ2DElyKedlw"
        "VG7r3LJ15/a7uwkiAkBnBveH5Vp3bSk66mB/Le90VJgS4HONru2PPbx7h+UY"
        "I+H5CEMRrOysyqpHosQzs9zpeBJdl8jxim2Aax9X1w++OoKrDImr74SEXcAo"
        "UBKeizEUi/LQjq0bevZ/pl8XaWDwQFS+x7+y48QpAT7XcF33kccfa+3qM6Wi"
        "LBuYjYQBV6bhKqwk8dKAqxT9ytxc27hUx0XYrnmI3PSPysOArDwqdogBmQml"
        "YE1U3Ni788AjT3zGCunJCZ4efNhE9+yHNpBpDtZPNNy63eLr6N3zzKGhn8xf"
        "saXQzqHSxA0KpJIecK3Neo13VKpuebxmfGW6vm4cRVa2xa4TQ1zlJmJciXC7"
        "OiwIKIBIczEvmNb8ho5drQ/v7WjNL/D0hyxAwKjsig0BzJA4tUQhsgUBRAdJ"
        "3WIdODN4cPe2467/WaOLOK7L6RezXZcB3JQAn3c8/MRT41em3v/wXSpbnokQ"
        "tNRrQSt2JdQSKx1QWMlamRYBQEa+dt5MNPgbjVsAQLl+nmVYG9CDkuz+8lW3"
        "qRCDAC0ano2gxJ072776xVaWcG7stLAFZhQbWUIRawUIjAUAsIIcAyNYQBS5"
        "hceHGZSiM+cP7tr5kafsZ5H+jPtSXHQ9iUScWrrRpCrQb5Tcmcbnv/KV3XsO"
        "Wo6laHkqxLlYkqg4FECp2Ly4sue1oiAl5qwQr3Z6VscZbjSOAmvHK48VbQlW"
        "P5gFuJr9hQw0b2Q6wrIISUtjBlabEPcIAurMmYNxrOiu4qtZ6nx6qVyuY6aV"
        "ret0BfjcI5Pv/MY3vxGaePD0x1TSMhMTs8k5olCYCQgAeEXWUKoey6SHClyj"
        "G1XHWW48fn22OCLeMIccK3M/QAxq0ciCoQhE095dBw8+sgWiqTVzVHxvGICk"
        "zpx/dOe2o15wZ+uAmGygXoriTHLe6ykpfv1Xz2ts3vj7v/97u/Y+xsRSimXO"
        "qpkQywYJGLiqlCf7tAQVTiQWc/K0+rjT8UoERFLxapXPJ9n7ZRQsMk1GMmvQ"
        "oLFRa3v3bz/7bKaxRa6l3r1bBwTPnn30jjINOHZR/jSymXUpHg9E+cj6/MYX"
        "v/1fPXXoKZ3x46USLgJORnregBWLwlWvEFT+hOSPFadnZYN27fgaACeTIt9k"
        "PHmWHApELDFYcRYsTkdYrBRvFLb5pnymsREYECUJ7r/nbfIEVHvL4B0dlpyo"
        "ruH1uBisdi6lXqD7CSISNHV87Rv/dXtb6+vvHp6+POTGvlirSpayygRKNICV"
        "lVC4FX0oCdpZ/ZMn3p7rNeCV8eviHxAwsQSwOuGwEbXEVLJSEowRmFVWAxKU"
        "wRiDHAMBY0VE2VQ9OfxZC50jMFBbbrCta/pOP7q8eLKlW6bHX3CcEABilaZE"
        "3ldIAtB0pvGxZ393Y9eG1946PHD203Cx4EYZKBsnY9kn65PVQlDZuxVZFcIG"
        "1859Ijcex2tj41Y0GREhgFh0KFhkLIuKkK2Ah35ni97ghhcWebrilSJiYEYl"
        "YpgUswUgAPOZZwHA9voLnZsnGIHoDqsEaVhY6m/pgOkrL2httDEpAe5XdGx/"
        "5Fsbek4e7Tp84tTo0FlZFh37WGTHZfHJBsgOJiGgCAA3z0mUGgYFKtGkFIoO"
        "BSOGsCL6lsBpzfg9DfHjYIuCFxGrC0simgaAgZlQ2XtR0ora6oc2dE+ABbjb"
        "6Xthrr+lExYu/a6zjuThgSMAAGTq848/90LPtl39Hx07MXBhYuS8lK3jBlJm"
        "KYjjEztoHWQHRQMBCN9BkyGpliQyQLGoWDBiCQUjUBbZCmtRed/vaLD7sLQl"
        "No4EU06iV139GsOCAgQkhg19RpcjIrY2XuzsvvLZhWVhvr9hIxQd2M0sAAAd"
        "HklEQVTNvpQA9z1au7c9s3Hrzn2nz546eeri6MjQuXBxyfV8LGl0gDQoB8FB"
        "cZAVigZQN9rnWi1nDGCFDGAs2grHArGgARWjCIBl65HTnPVac3YrxJvA+KyW"
        "lO+grFTbTAoAoVilrNhkc5kQP9P8L6LaGi92dE3dq/u2tNA/XtjS0jKdEmA9"
        "2AYbenZv6Nm9+8rQ0JkzQ8OjF4ZH5yZHzGKktUtKg0OiGQlQoRCQQiYQxNU9"
        "F5OQOEoKOqxUOQeD2kBF7l3SOddtyGKLa7dA1CGCSBHoiICBQVSESVjo1c1c"
        "AwBMwIwMTAIWCO9ia0xEtTdd6Ng4zXzvHEoayK6f6tAPNAGqaGnf0tK+ZV8Y"
        "Tl84MTw6cfnK9Njk1PTU5bhYUOAlm8IrXkQQWBMvjwIoSBYoic8RQSsAwFml"
        "Mx5lPV3vSzvaTrANQgagDFjJThNgALPGwCUiEiMoYKu7BgwCd+ECElDtTRc7"
        "u2udqlnVWYuOWrrN+wwkvqCUAOsNnud17XqiaxdwaX5yfHJyYuitD0+MXjyj"
        "XE8cFGvZGI6Zk5yBJLIfUXsueR6DoEbQjva18jX5LtSRNKNtlSiLAICx4JIk"
        "vdaTzEwEZGCMMImFhmpsEgAIIwpbQKpUtrsLq7e18eKGrprmfiIQCGL4Q9dl"
        "tn+jqXg7ClBKgPUMCho39DaWlivdX5AwXCps2bH38Yf3RmEUR6GJjBUBBN/V"
        "p4fGLp0/TVq7LXXYF7ALtpE5hwDARsgAFLmSCIzVil1S5Q8gcCwrSTdgjXLr"
        "fMviIouuxGYkFnLti4AItueHOrtmWKC2rEjt+C+UIz8mcei7LH91Gw6sHy9o"
        "SoAbYXFq9MgH77535MTSzLhfl4nD2G+o/9ITj+578tnr36xe+dHgqY9dR2Pe"
        "j/sESwIMspQER1ezxwTWRhZJ9V8BFZNlACRERAdNZm9zR3F+4iQlEQucpG7W"
        "Kv0I2J6/1LFhpsaSuogKnW+VwxZAYCYjvnJe4vivybk5B3RqA6xTlObGB/o/"
        "PvzxmaHBU2xMIv0A8IVHHt/z0MM3ng3DMhuDCtGAFASj6jTMVa3mxkFylXEU"
        "U1F7NJLvagCA/OONYOZG+5XCCgdqs+kRsTU/0t42wwx0O38/EYgodL8dmfxq"
        "ncaYjOd81/JfK7pJ/ROTEmDdYWlu7Oynn5w8c2HwwunS4pKbybiBjosl8tzH"
        "Dj7x7Je/TMGNW1REMVfUGxSIBaJqXWdMgv9hlYdfKpJayYavREZbEREktVQu"
        "TVw407x5R5BrwvwXmgCmR04prDUwTgTbmkY6Omdr3DizrNB/MYzz10ZICxjj"
        "6+AlsX99wxpA6KSxQOsI86MXzpw9+enAxZGLA6Wlgpvxg1zWRKZUKDW0dT9z"
        "6OFHn3rOr2++uRjZRN5BBKysDl1OKlKvjo+o7C4nNeGqHQAMs1jtuhMTl3/4"
        "6hvPfWF535NfFhHMf6FNeGr4dC16vAC25Yc7O+Zqkn4LiAq9b8WlJtQ35pIx"
        "GVd9j81fkbt83cvpCrA+dP3Z0f5jx06cOj9yaSAulZxM1stlbRSHhUK2qfmh"
        "/Y8++fi+zp2Pq5vHfllrLVsElMTMNWvUFa4EDl2rqawYthU7gSygFQQqL8wX"
        "Z6bntvetSKFw89MtYKcuDdzC75Lsl7XWjTXk52ry+QiwaHS+acpNt35vbAJF"
        "f8rmr0hPX8OflAD3N3h+vv/0kfeP91+6cNqUyzrIuNk6G4Vx0eZaOw4e3LZz"
        "d++OXQ+72dt05gqWlnhlOhQRLiPFV2PMkW8aMSQVYggAshZq9MU3Oqlm3eZU"
        "bQZmpuZnWgEmhs7eLEFejGrtOd1pRpcWa/L5WKvA/WZczhPePhyehcD+EcBf"
        "kHuVA6JSFeh+xuzIhbfeeeP4JydKc3NuXZ3yg6hYVFpt6N66o2fz9u2berbt"
        "u5nGf+0ciWgMq4qSzxiTsJBFIVOtnrWmjlCSDZOUFWUQMBCC0QgdgqySlMsr"
        "m+avWWSONv2Ph+D/mLh4Fq4LSmLRbZv7g6YxFgKyt7UWhJTobxrTVPvOgkIB"
        "/h7Hf0nOCgdSN+j9i9GTR19+5ZeXh86A0l59LioUvbps3/5Hd/R2b93e29q9"
        "R91JsHvsiLGQ6D9JXRO0yGi3mAN1+QwRoIAD4BC4BC6RC+w67AO7DueUbGn7"
        "1hMDzxVyxSNPHF2rgq8R5Eflg/NN398m/3ns0uCqiDkQUO1b+t36y5YVALgM"
        "WgEAJC5TigEUUKWcHZABJte6L3DceEcBoWyJFJP9HsNfkjMNGlDblAD3JYZP"
        "vvc3P3plemzIrcuAgI3i7fsOfuHhvd19O+6yAqENxVpYyQUDK2BZiFFKWkgJ"
        "KABF4CrUBEqAHAQCJrYgReDB+b8vHJq1XhNZwhXc8Is28+BI0x904d+NXrhQ"
        "1Xzae/tVdpwt+Y75NNaHx1qmy0HgRXt3XlGOTUL3qtUZFSKWezJysc4P8irX"
        "YJuDUgNK0pO4BpWGgOz3WP0l0XRqA9x/QMSJoU/+9ievzoxfcnwvKha11gLY"
        "kMnu3H+wRoXnBj5Q0SwrTn1msQAWRcQBzqLVCpCBiFy0WkApdAA1gAJxiLUI"
        "GJm7+PNNu59bwq7b1llol3Hb+ExnH4yevQBC7dv7vboxY8gl859OOG/051FJ"
        "k1faeeBKJbBipfIKEpCgjTegKiyY4oKBy8QCjBoUNNZRUO9kW6i+CfL15Twq"
        "giQtiNcoXAKIIBR/n+X/SlMi7z8UZsd//srbM2MXSSk/l9+1afPY5fGZiUsf"
        "nzzWXJ99+nd+z3XvptKTrKQ8VnR9A8yWhTWxj6JEiICAScjVQCQOsaNFISoQ"
        "x0kaLcmVs79o2/07BW6/LQeYGeu/1LrZA+dHXm6MmQDg8LjzxuEmVCKC23dP"
        "3sBFKchmA+JqvYcQCAUZCotRYSmeGUGbVMdwTUPOzzR7mTYn10gt9cvtSjEx"
        "Vs3levp90zwB0J4S4H7CJ8cOnz/7CSBkGtq+/ttf2n3w0GD/Rz9+5ZczY5fe"
        "Pnqka3PHjoef+SzHFxEWa00S4IbaY58MIhKRIlHIhOAoUQq1sFLooEUBElCu"
        "ZYbpk//cuu93C9xx2y+6dGnzf/NH/8Prv5pFRAHJeuZHbzUksr2jeQHVDbac"
        "JW4HvJ27RxSBQkQrpflSab40c54YmAVFq4bGbNCmM+11dc3Y6s60YG79iM0D"
        "QYDF2dFTg8OJd//LX3zioS9+jYh2P/HcUqn089dK81cuv3v40+7N2zL5zjvW"
        "rEAwab+RqEAGAUCANYCDAoBErAkUgUIhAS2iUJQwASgSUEAMCILC8/0/adzz"
        "wq05MDi448//9KuvvT7vOhAbAICMHw3MEzogQk1ti0nvmdXSb+P2tXP/rfX8"
        "q9sYJApQMQFzeXapPLM8e6Zi6diMNG0D1dXubHB62kdfuq9l44EoizJ2/tzo"
        "6BCSemjv/ie/9CxRpbbZoSeffmTfPtf3B8/3n/zkxN2ZFonrHQHYWmZjjTFs"
        "HLAawdPiIjgkDokmdpRoxYpEEytlAUAljbwJAMDEMN//kyyN3+y73njj+T//"
        "06+++face21O7op8E1+j+dyR9Neg7xGCJvTKUXT2VPSrXxX+v3/55H6XjQeC"
        "AFPTMwuT403t3Y8c2Ad+I1Qj7/3Gp774zKbe3cvT0+fPDxcXZ+/0yIoQV4J1"
        "pFryn0GBaAdIgVasERy0jhKFVgtr4OTbFVqCNSFrVQ5cbwy8Mfz8f/zfdrzz"
        "3mwlPO720orAbb+u33eFaFphSoDPO0RkabkQlku7tvW0bztwzavNXX2/9eTB"
        "hg0dg8OXrgyfvtODu5ocrSqtHVfEgkUUigOigR0SRZYQCNhRgiBAlSigG976"
        "G64DR48e+o9/suPNt+d0jbO5oJg2FgUpUgJguFAKY6WdDa0tnudd/4a9Tz76"
        "6IGHZi5fnBi/40Rvo+uUo0QEWQSAEZkAFBJAYvUiiKOEqFJPGhUSAJIovKkn"
        "0cQwc/KnzXSpKv3/4d8dfOfd2VrdVIJi2uAm0o/EVG3dh9VeTQ+GJvBgGsFl"
        "8EWkrimfb6q/yRLR8PgTh070n5mYumMCKKV85SYWhQoFwYIBBMZGTBwyiYZT"
        "UXVqiNFPYFmGT73etvt3fvXBN/7Dvzv4znuzSt9e2bAACm8i/QSlCVyYc3ta"
        "urd3NNebvDIe+mHBn5+EyeHlCcYSXO+EJaCkhunaZDRKCXAfwfd919VepqGh"
        "4ab91jf07P7CwYfmlkpcmq9lUyyO4+mRs9NTk8tLiwNDI27gAwMUmYoADBgL"
        "9FWEXt1dkVthYZg88S///n/6X995b97RYmwtqY3EcTPeKM7hdL/qxpZv7don"
        "7OgYBEF0jJZypZZ63brD2X8xe/LswiCv4oCgRNOoAifX5JXCUCBel+LxQLhB"
        "s66XCXxQt5LsfQd2Tc/MstxmeosLcwP9x86dvzQyMTUzfaW0MEdE2vU5ZopB"
        "WJCQ+R5slEaxu/nAR8eOzziapKbCXFakXSmVhGasxqlTug0afmfnAcNEuNLT"
        "TAEIMYECAG16y3ua803vzR2peo3E0PFB89zu3m/So0rjsab3+6eH7r6sXEqA"
        "3yDqmurr6xuVcysxauza3dReAvdWbdmX56ff/sXP3zt+uDgz6zieIuV4gTgo"
        "Gkk7opAUIpFyUOxn4oBYtXHfJ9mmS9bWLnAEqJiIAKocUIgLl9Syke/u2R2v"
        "KulLQKFantKTG+NeBQoAEKWx2L07P3dq9nzCAVsSYO5uzpsgAp88reVqi6iU"
        "APcVNja3jDVesctzt3iP53lwIxP56p2an3/r1Z+9/stXEdDJeJQLKKPRU+gS"
        "OkpcAAdFAys0LuPi3fsHBVXbnrO55gt3Iv0rNq7Iag4YA6dm8Qst3R4H1blb"
        "BMez58+V+mcG9MuLF7//8FMtTpYIALkn3DegxpnLqDCchsDJbGtqwDIyrfF4"
        "pjbAfYa2nQeDsxcaOrNzEx8hWdfvtJirVv2vdD4VsdbyWlQHrbUi0rPvsX+9"
        "/5BhkHDp3NG3S3Fc6RocMRhgTspioYjYFrlr6W/ZeTbXeg4RiZj5NsKWz1Ws"
        "37W2ABGAiC2MKSA+sLFTjFTjqI2OBoqnEJzmXpw5aX/6yYk/e/ypOGnsjbCn"
        "ceOncwMAenpRWrTXvNQQOknrg8rHk23EdSMbD4T7y3XdrT2bTMSEoN0tohqS"
        "wGMiStSC6tME1ZeISClVHbTWmqhswwID9R48pLU2ABbEEBgBi8IIFsTKPZD+"
        "Wt4fBLhHvQhQvpE9TEqpC8vY5zR5cd0qlxdeUIO04ibalJeReG7AVLYdWGAz"
        "bCkMUThJfa2bnzy4vexV9B2XiNAhrcFbV2bAg+L/7dh+wNpYed3oBPfkgALO"
        "9ocPec49KxVelf4ai3j6PjW3/Hn5L/77m73BGCqX7cZ8kxW7+veet0vVxGGv"
        "VcDw5fmrae9u7MZL+Xgu+MOeg49nurlSss5uW977/dbvf6fvTw4daBWxKQHu"
        "v0XAy3bUKP1JIbcaRNbZceCJOs//Ly/9no9N+X9zYeShW7yHlwk0NbdnV//I"
        "ghLbsHp16AhoWp4vVFNiLKum5iyLxHQ1BBoIJBNIptUpNTvaXU+C8aAQgJmZ"
        "6mt8s43DGpUQIadn/5NZN4DPEBTDMd2R9LseZf1/c2Fk/3W/5ZrPm1AAZEOu"
        "1XUyq0sL8VqVKaeoUCjLilIkjJ09rgUzDqVYGQAQEscJjNcGToF1eZ0JxgO7"
        "BX5zmTYFKA5x+bJIrepN7/5DGce/a+nv2DdwB3O/h2j/ZGgkr/jaEs12bXgF"
        "RwAAWahTXs5zs1UKMK75oKdwIY50tVYj4eZG39Ly//7mP5yPhgnZQ0/c7svt"
        "P/3rsf/8g6n/+/2Pxu9hhGlKgM+Z9EcFLlwWUCYuSzQCNXdE7DvwZNYN7tRB"
        "LqA6dg/kmmuVfiRl4/92emYTMpryBGLppr9r0gcZwAcXALRf7wZ1wEnywZo9"
        "XZewJCFeXSLY54zvAoDnAHkUcHajGLISmzgMw9DwetoGSAmweso0y1QeBgWg"
        "ABFNXLblcak5prJ3/6E638eaYx8EVPv2gVxbrT4fYcxlvjU3ezV5X5bGL2y6"
        "GjeqVv2YLGgNAAAZndBY6aybqbdo4NrAHjFrt7Y0+KQBQJSvbaZrPTk9UwLc"
        "1OoFu6yjYVBrFOk4Ck35MnGtd6l376HArcnOZtbt2wdq93giYl32OyPDa5pz"
        "MemR/Plq7vo12cCVZ3xVkVM662T821r8DlBStYvr9Q16wqYEWH/SL2ZJ8aUb"
        "vSRsynE0XqMQEMDW/YcC/zYcEKvatw3kWs/VeP8JIee9ODa859pxNuoOS5Qo"
        "z7umeCJWunasspKZkshQy+tfPFICANhl5BG+edq4ictxYUpqKJiVVL3dtu+J"
        "bO6mURUs1LFjoL7tfI03HxGz7h8Mj++vwUK4wU9rro3i1ESEK9eCiAxItKb7"
        "hs5UWtTLTaJ+0p3g9SX98aXbTvDWhHZpvPajdu58Lpdx+LrZmS119A5kWwZr"
        "U5OICLLuH1y6fKAmI2GVXBJKYhKIs+YkVOSLCBAhVWwDQVCr+tg7ypPmFZfo"
        "A/D7P9AEQLsE0aiIsrfrPoQKLcdhodZmo0yU2fbthpxjV1cyRNXeO5BtH6zR"
        "50MErn5pePShWjuzs4EVTYYFSTMAhGudnoSI4ggDKIQk6lPABUo60GjHx6Ar"
        "okVMVgBOCbBOYYyLvBQXrxDVHL1PIBzGxZp0oQRB34sNWWdlDVEbtpzNdQzW"
        "Yk4gCjMF7h+Ojey70wuDFTdlEv5dguXrLAqvyjAkHRnOKl9QlOvpYIMSKcUF"
        "YQKA1fPCmtMWTAlwf8OhGVsaR0S+QzvP2tgW7qDlutf7Yl3giqj2nnPZtsGa"
        "PSoUuH84PLb/Li4NrU1aCpMLADgbF9aqSajIvWr/IpYBM66rXM8N2pLg1uW4"
        "kKwAkBrB61TzWYiX78DBv1bRJmtsWJyp/ROl5j/rOzhwR9Kfdf5gdHi/ju+2"
        "DHlsCUXlBAzPLl/b99fzvNXmcjEydfk6CTqqjSuvzIVIBAxX+2Ere7Xjn5WU"
        "APe39JcL04yfIRFCgE0UlWoqIjQ/3/LtF/6toaFaPZ4EPr10cfThuzqxVfZG"
        "ZJTLytNT43PO2piOhmrWm6ApAjA3tbYCVZr7WQznCoWkC5is7uAtDqAwA0dX"
        "R1HSukD3nfSXpuleZAGxjaLSLN7SIRhFmW+/8G/ffuci1hYrRwRavj8yeuCe"
        "XCwbszmAk4sT5K+JYNvsbBYJAQAJihOAju5rulqPMaqfLJUXyUFgsSu9WUWk"
        "jvMCBm0sYqsLiEVJCXC/Sf+9u2K2Ubl403Vgaanha8/9L2+/c1HRHUj/2Ni+"
        "e3jJ+Q0mDuH48qk1K8Dy1qzXmjh5Bmbt/roNXfHmlRldDfMJAa0CBKAoLlbV"
        "tsbJ/U3Y5Re6D7bvYInXjfQ8KARAxKg4Qff6cplNXJy5Xrm31vvGV/7nt94a"
        "0gqwhmlSq3sh/dd9kZOVtgx+1H/RW3WGsciTwbNaBedOLu7KNn390PNVV6lV"
        "pZMj4wQEShzHmZmdQ9YAQLaMi+O/V3/ou50Ht2fbr65797+f9AFqkPHrOKwg"
        "GjZcmPGCJrkq/U7vpv/n7Xcuep7U0rYREbV8d+Sezv2VMxHZtMUePVN4d+HD"
        "x+qfRmUBANHqUsOz/u89/mShSTUAV8KfCPRp7+fGhgqVUtjboE/3z6jeKLKG"
        "o1FFaFkYVUQRr6P5Mw2FuDfrQKk0l0SVWetl9OvDF07UUssNAETQVS/dK73/"
        "eqhAHtqIvzx9fphOVlO8mBhAN0iDrNTbInYv51/vHx1SqBKlv2k3tdYX/s/X"
        "/+J8+HbBXyhl5sdy5z/Uh3984QSBm64AKdasA8BxuTDr+K0Z/VpY+EAaasqP"
        "IQLfeWl0eP89WZ9uNpl57fZhUn/7/rHnd5d/K/9oKHRN73lR8QnnXy6cH0Vw"
        "EnNdEAmgfpviaPrw1BSCRgIRKwQoBOsoISYlwD1SsUSsFV1+JbRHoeZIh/rs"
        "SxcH9/3aT47BbbWP5XFiov8/DY5+4dEte+Txq+HSuvyzub9FEaSrdRxRBBCZ"
        "WRQCIIgwCCAi4zUqVqoCpQAEseK46jhJrdKvNezf88K5sW/SvZuDhG7FAVCY"
        "2aLzdcV/fPdEOTdyVQKMn/XrbjipExGiWt8ikhLgXsywVrt03KFjtd50hJ07"
        "vv3qO98TkTjouTccuGVIkwAhabaS6RIw8vGVS86qartP538rCAIUuUFIOBHc"
        "fN9E0n2AdO4XVq467tAx5pqMQiLYsevFX73/4ooRLHHQo+5F0dkbZjRIEv2s"
        "K0KsPWmrc/r7x4JVU7633Pav3O/9dt9vWy7f0MLBm8T/ecpJCfCgz/2aTrjq"
        "aK23m2Dbjj98+4NvXSOjoX9vOHBDNYYctWqdwM3bcbS8/NryYbUqGsqidU3d"
        "TTNgFKobyXpXj0oJ8AA7f1j5zglPHWGGWjITEXHHtj94/8g3bvjqr4MDRCSk"
        "rwnrVz48upNeO37irfn3fQAyPrBLxle3jg4kuIYDlvm3dnw59QLdJ8J6z/0V"
        "Qp57Gu2HK2IFwAAEJMAChJGwz3h1F0wp2br9O+8e+/otDhn6PV75ooW7LTxI"
        "a3QwQsVImKhGspq3oOvhsUf11Pypv4tOt7Q3udqNTVSaLCsMbn180pqNIYVR"
        "bP/odx9uHno6JcCDQQAGESQiAECFYLTnfgzmPUve3EIQxRlUTfVZP9PQWA47"
        "PX9bwaLjvVeXuTB4wXZvHHfd6MBD33rj3d+9rb8/4YAxdxEIjUircrgU8u22"
        "oN08AcB8cR4LAlaYOcmNR0ROCgZTJcuZLFi0YGOkTFNj/sCjemvrU+tA+uEB"
        "DIWoFIJ2AWIWUeQyGUGyLAp8oyyUluKlQrxc4lIpDKM4Co1htlasiWNbqZau"
        "lPIzXYj/Xa4uqG/MtGSzdXV+JuMGQeB5nuM4RBTLjpkldrPUf1GXlqYOn5hu"
        "aJiM4xgAurq6nFVVda8hZ+j3OOYi33UyAAAoFNLViwUBNgaBxQoggiBYCwoF"
        "kYiYCJQCTYq0IKJCJEAREQso+ab6jZuguyNoMVv9hceyw7sxU3YKERTWiWA8"
        "KATQWi8uw8mzF6an58NyGIURA5dKoYgwmygyK9tC4LracRzXdR3P8zzPcz3P"
        "dYJM4Hmevwqe57mum4i71lprXa2ovsZP4kFTEzC3MOeNMWEYFgqFc+fOLSws"
        "lMOSo926+owmcpygznVbOjuTPpax7mlRZyMUU45DcSw5NTiYBJPiDihiREyZ"
        "ks05QUEkrdHV5CrUBKhAITkIiEjgElgGFivWNjTWdW7GLV1Om9rsyrb82BOV"
        "Y1dz4DoX15tgPCAEUEqJ07y4eHZ2fjEpt6Act7ExIK0dx3GUqgi947jXofqS"
        "1loptVrcVzcZuI0rhkhr7XlefX19W1ubMSaKonK5XCwWl5eXZ2ZmLi4uLhw+"
        "vLi8LDZ2XTfptOEFnuu6QeC4ikhrRynHVSCQCTQAGANZg+g5wAAkZn4ZYxRN"
        "oBxyFGgHlCJHASpwkBQCImkQBhbLxmZyQdcW6u1y2v2NCje3XXq2crrj8ODg"
        "AVKBcrlcX19fe3t7siBorR3HIdf1VmR6NZK+GEoppVS1d0YiyvAZYkuTDyZf"
        "4XleNpttamqy1iZ8CMMwDMNyuZz8P/kjDMNiMVyI4zCugI0xhplN0t7mC+7m"
        "93FUBCmTEdCoNSqFSpGDoBAQBa0Y62X8rk168ya90etQ2NM6/FzlnP7/du3Y"
        "CAAQBIJgRAPff4VWYWBq7OCwWwOXAGv0Km/QL1CSJK2aPIFVPTwnzR736+YM"
        "BAACAAGAAEAAIAAQAAgABAACAAGAAEAAIAAQAAgABAACAAGAAEAAIAAQAAgA"
        "BAACAAGAAEAAIAAQAAgABAACAAGAAEAAIAAQAAgABAACAAGAAEAAIAAQAAgA"
        "BAACAAEgABAACAAEAAIAAYAAQAAgABAACAAEAN/b0mZho3Q2iFYAAAAASUVO"
        "RK5CYII=")


def getErrorImage():
    """Generate 'pending' image from embedded data."""

    return base64.b64decode(
    "iVBORw0KGgoAAAANSUhEUgAAAQAAAAEACAYAAABccqhmAAAAAXNSR0IArs4c6QAAAAlwSFlz"
    "AAALEwAACxMBAJqcGAAAAAd0SU1FB9wDCQACMfBoIRYAAAAZdEVYdENvbW1lbnQAQ3JlYXRl"
    "ZCB3aXRoIEdJTVBXgQ4XAAAMd0lEQVR42u3b23Pb5oGG8RcnAiQIgpRdUbJ8jONNk4tM053O"
    "XvZv701vsrvqtJMeMtvYUZSKlHWgxBNAEOBeGGKhCKQkS4oZ6/ldaUSK/AYiHnz4ABrHf/3r"
    "rBfHAnC/NF1Xdi+O9f3JCVsDuG/CUCZbAbi/CABAAAAQAAAEAAABAEAAABAAAAQAAAEAQAAA"
    "EAAABAAAAQBAAAAQAAAEAAABAEAAABAAAAQAAAEAQAAAEAAABAAAAQBAAAAQAAAEAAABAEAA"
    "ABAAAAQAAAEAQAAAEAAABAAAAQAIAJsAIAAACAAAAgCAAAAgAAAIAAACAIAAACAAAAgAAAIA"
    "gAAAIAAACAAAAgCAAAAgAAAIAAACAIAAACAAAAgAAAIAgAAAIAAACAAAAgCAAAAgAAAIAAAC"
    "AIAAACAAAAgAAAIAgAAABAAAAQBAAAAQAAAEAAABAEAAABAAAAQAAAEAQAAAEAAABAAAAQCw"
    "2mw2wWr4/dOn85//sLNz7ccBAvABd9r3xc4MAgDcwawJBOBOLfvQXffDyQcYHwKLgAABAMAp"
    "AH7x57OmYehxEKheqch3HLmWpWGSaDCZqDMcqj+Z3Po412s1rVWranmestlMg8lEPw4G6kXR"
    "pa9XtW1tBYFqjiPfcSRJwyTRMEn0Y7+vaDpdOoZlv+O0igDcK03P02dra/Ls8//Whuuq4bra"
    "qNf1utfTbr9/a+/52dqaNur1c7/zbFsPqlV9d8l7bdXretFsyjLPT0QrlqWW52nT9/Vdr6d/"
    "DQb8cwkAlml5nr5cX5ckzWYzvTk50cF4rHg6VcN19arVUtVx9LLVkqRbicAnzaZanqdv3r7V"
    "SRzLNAxt1Ot6HoYyDEOfNJt6OxopTtOLO38Q6NN8LNMs0/8dH+s4imTkIfu01ZJtmnq1tiZJ"
    "5yJwdmTnKgBrAJBkSHqV70yS9M3BgXZOTzVKEqWzmY6jSNvdrqZZJkl6HoaqWNaN33fD9/U/"
    "nY4OxmMlWaY4TfX9yYk6+c5qGIaeNBoX/s4xTT0PQ0lSNptpu9tVdzjUJE0Vp6m6w6H+1O0q"
    "m83ejbfZlG3yUSUAKJ9KB4Gq+flzP451OB5feE6SZdofDiVJlmnqacmOeV0/9PtK8qgUdfL3"
    "kaR6pXLh8cdBMN+hO8OhRkly4TnDJFE3fx3HNPU4CPhHEwCUabju/Oed09OFz9srTKPLdszr"
    "OioJjSSNCzu0VzLT8AvvvV+IxU91LwkJWAOA3q2kn/ni4cPFpwqG8e+dMJ8x3ETZCr0kTQqz"
    "grJTjVphvIte46ePVW0+qgQAlwaguJMv/cffwjl1mp+jL12fKBlPMQqTkgXCssfcW1izAAH4"
    "KMVpqlq+Q/9xd7f0vHxVx1uxrNKrBD8NRbwkFGAN4F4rLqLdxtT+ro0LU3tvydS++Nh4yakC"
    "CMC9Vry7L/S8lR/vsDDe9Vpt4fPavj//eVByB2NamOkYfAwIwH31Y7+vOD9CPgmCpUfVVbDb"
    "78/vSWj7fukCX81x5gFIsqz0xqXiIqHLIiEBuK/S2UzfHh1JeneN/7fttp42GvIdR5ZhyDQM"
    "VW1bTc/TJ82mfre5+UHHm2SZ3pyczMf71caG1ms1VSxLFctS2/f1VbstM19AfNPrzYNRNCic"
    "+jxrNFgofA9k8yNxFEX6y/6+XuXfBXjRbOpFs7nSsxZJehGGckxTn5dcvkyzbOl3AV73enpY"
    "rcoyTW3U6xe+k8DtwQTg3kXg6709bdXrClxXNdtW1XGUZpkmaarxdKrjKNLRFb6l93NF4Gg8"
    "1lYQyHcc1fIFzFH+bcDdBd8GPBOnqb7e29PjRkOh68q1LDmmeeVLoZCM19vbs+/z6RiA++NZ"
    "GLIGANxnBAAgAAAIAAACAIAAACAAAAgAAAIAgAAAIAAACAAAAgCAAAAgAAAIAAACAIAAACAA"
    "AAgAAAIAgAAAIAAACAAAAgCAAAAgAAAIAAACAIAAACAAAK7LZhN8/H7/9On85z/s7Fz78Z9j"
    "DD/364AZAEAA2AQApwD4iDFVBjMAAAQAAKcAl/JsWy3PU8vzVLVtuZYl0zAUp6mi6VQncazd"
    "fl/ZbHbhb/9zY0P1SkWS9O3RkfYGg4Xvs1mv6z/W1iRJ/clE/9vp3No4ztx05fw2xnDhyGMY"
    "Wq/V1PI8NT1PaZapP5lofzTS4Xh88yObYehxEKheqch3HLmWpWGSaDCZqDMcqj+Z8CEnAIv9"
    "16NHpb+vmaZqjqO1alVbQaB/HB7qKIrOPWdvMNCrfKfe9P1LA3CmU/K8m4xjFbbFIr9+8EC/"
    "qtX+/QvLUtVxtO77+uH0VN/1eu893qbn6bO1NXn2+Y93w3XVcF1t1Ot63etpt98nAOzq5YZJ"
    "ooPRSMdRpDhNNUlTebat0HUVuq7WazVVLEu/fvBAX+/tKcmy+d/uj0Z62WrJNAwFrivfcTRM"
    "kgvv4TuOgnymkM1m2h+NbnUcq7AtyrxsNtX0PP398FDHeTBanqdPWy3ZpqknjYay2UxvTk6u"
    "PdaW5+nL9XVJ0ix/jYPxWPF0qobr6lWrparj6GWrJUn3PgKsASzw33t7enNyopM4VjSdKpvN"
    "NEoS7Q0G+vvhof6ZH6Ecy9KzMDz3t9Ms00FhZ97w/dL32Cgc/Q9GI01LdpybjGMVtsWiWc92"
    "p6PucKhJHpTucKjtbnd+GvE4COSY1/t4GpJe5Tu2JH1zcKCd01ONkkTpbKbjKNJ2tzvfzs/D"
    "UBXLIgC4vu5wOP/57Hz/3GlA4fG278so+bC2C1PgZacJNxnHKmyLsuePp9MLvx8liTr5a1mm"
    "qceNxrXGsRUEqjrOu/WUOC5dS0iyTPuF93h6zffgFOA+bRzT1Fa9rtDz5FqWXMuSVXJUqtkX"
    "N2MvihRNp/JsW45l6WGtpreFWcHDWk1OfvSJplP14vhOxrEK2+JCAEpOdeanT8OhHuUzo3q+"
    "M19Vw3XnP++cni583t5goEdB8EGjSQBW3Ga9rpfNZumH/KecBdPIzmCg583m/DTg7YLTgs4l"
    "i4Q3HccqbIuiuOTofyYqPFa9ZtCKz//i4cPFpwqGcW4dhgDgnDBfLDIMQ9lspt1+XwejkeI0"
    "VZKmOrvYVby8VhqA4VDPwlCGYaiVHznjNJVrWWp53nyhqlOYQt/FOFZhWxRN0vRKj133/LwY"
    "gOJOftnMhgDgwrnk2Qfon8fH+lfJEdq9ypEuTXUcRVqrVmUYhtq+r53TU234/vz1z1bW73Ic"
    "q7Atiip5CBc9dpVQLNretXyH/uPu7p1cEfnYsAhYIiicFx4suCnlqtPT4tH9bNW/Xbz2v+Do"
    "f9vjWIVtMQ/GkucXr92Pl5wqlBkVLrXe96k9AbjJtKgwLbQWTCWLl/CWORiNlORHsqpt63kY"
    "zneYJE3PXS68y3GswrY4s168AWjJY4OSeyeWKd7dF+anWCAA11Y8krRLruE3KpWlH+Kimc6v"
    "ehcvO3VHI81+pnGswraYB8P3S2cNNduexyTNMu0uWckv82O/P19gfBIEF+4EBAG4kuIdeVtB"
    "MD9qe5alzXpdX66vX+sOsuIqf3FxqnPJtf/bHscqbAvp3T0SX7Xb8zsIK5al9VpNv9nYkJlv"
    "n91+/9rn8Olspm+Pjt7NVkxTv2239bTRkO84sgxDpmGoattqep4+aTb1u81NZrvs7uVHkgfV"
    "qlqeJ9s09SwMz93h1hkM9F2vpydXvIlkmCTqTybnzqf7cVx6e/BdjmMVtoX0bjHRsyx9vuBS"
    "3Q+np+91G7AkHUWR/rK/r1f5dwFeNJt6kV+KBQG4sj/v7+tRva5131fNcTRNU53EsXpxfO7O"
    "tysf9QYDBfkXhM6Ogh9iHKuwLSTpb4eHakeRWp6n0HVv9duAR1Gkr/f2tFWvK3Bd1WxbVcdR"
    "mmWapKnG06mOo+jOvjz1S2K83t6eff+etQXwy/UsDFkDAO4zAgAQAAAEAAABAEAAABAAAAQA"
    "AAEAQAAAEAAABAAAAQBAAAAQAAAEAAABAEAAABAAAAQAAAEAQAAAEAAABAAAAQBAAAAQAAAE"
    "AAABAEAAABAAAAQAAAEAQAAAEAAABAAAAQBAAAACwCYACAAAAgCAAAAgAAAIAAACAIAAACAA"
    "AAgAAAIAgAAAIAAACAAAAgCAAAAgAAAIAAACAIAAACAAAAgAAAIAgAAAIAAACAAAAgCAAAAg"
    "AAAIAAACAIAAACAAAAgAAAIAgAAAIAAAAQBAAAAQAAAEAAABAEAAABAAAAQAAAEAQAAAEAAA"
    "BAAAAQBAAACsNrvpulIYsiWAe6bpuvp/KMHzybJ4Z08AAAAASUVORK5CYII=")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import os
import io
import sys

from PySide6.QtGui import QFont, QFontMetrics, Qt
from PySide6 import QtWidgets, QtCore

from VeraGridEngine.IO.file_system import scripts_path
from VeraGrid.Gui.Main.SubClasses.io import IoMain
from VeraGrid.Gui.python_highlighter import PythonHighlighter
from VeraGrid.Gui.gui_functions import CustomFileSystemModel
import VeraGrid.Gui.gui_functions as gf
from VeraGrid.Gui.messages import error_msg, yes_no_question


class ScriptingMain(IoMain):
    """
    Diagrams Main
    """

    def __init__(self, parent=None):
        """

        @param parent:
        """

        # create main window
        IoMain.__init__(self, parent)

        # Source code text ---------------------------------------------------------------------------------------------
        # Set the font for your widget
        font = QFont("Consolas", 10)  # Replace "Consolas" with your preferred monospaced font


        # Source code
        self.ui.sourceCodeTextEdit.setFont(font)
        font_metrics = QFontMetrics(font)  # Set tab width to 4 spaces
        tab_stop_width = font_metrics.horizontalAdvance(' ' * 4)  # Width of 4 spaces in the selected font
        self.ui.sourceCodeTextEdit.setTabStopDistance(tab_stop_width)
        self.ui.sourceCodeTextEdit.highlighter = PythonHighlighter(self.ui.sourceCodeTextEdit.document())
        self.ui.sourceCodeTextEdit.setPlaceholderText("Enter or load Python code here\nType Ctrl + Enter to run")
        self.ui.sourceCodeTextEdit.installEventFilter(self)  # Install event filter to handle Ctrl+Enter

        self.add_console_vars()

        # scripts tree view --------------------------------------------------------------------------------------------
        self.python_fs_model = CustomFileSystemModel(root_path=scripts_path(), ext_filter=['*.py'])
        self.ui.sourceCodeTreeView.setModel(self.python_fs_model)
        self.ui.sourceCodeTreeView.setRootIndex(self.python_fs_model.index(scripts_path()))

        # actions ------------------------------------------------------------------------------------------------------
        self.ui.actionReset_console.triggered.connect(self.reset_console)

        # button clicks -------------------------------------------------------------------------------------------------
        self.ui.runSourceCodeButton.clicked.connect(self.run_source_code)
        self.ui.saveSourceCodeButton.clicked.connect(self.save_source_code)
        self.ui.clearSourceCodeButton.clicked.connect(self.clear_source_code)
        self.ui.clearConsoleButton.clicked.connect(self.clear_console)

        # double clicked -----------------------------------------------------------------------------------------------
        self.ui.sourceCodeTreeView.doubleClicked.connect(self.source_code_tree_clicked)

        # context menu
        self.ui.sourceCodeTreeView.customContextMenuRequested.connect(self.show_source_code_tree_context_menu)

        # Set context menu policy to CustomContextMenu
        self.ui.sourceCodeTreeView.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)




    def eventFilter(self, watched: QtCore.QObject, event: QtCore.QEvent):
        """
        Event filter to capture Ctrl + Enter
        :param watched:
        :param event:
        :return:
        """

        if watched == self.ui.sourceCodeTextEdit and event.type() == QtCore.QEvent.Type.KeyPress:
            if event.key() == Qt.Key.Key_Return and event.modifiers() & Qt.KeyboardModifier.ControlModifier:
                self.run_source_code()
                return True

        return super().eventFilter(watched, event)

    def execute_command(self, command: str, silent: bool = False):
        """
        Run a command
        :param command: Python command to run
        :param silent: Silent mode
        """

        if not silent:
            self.append_output(f">>> {command}")

        try:

            if silent:
                ret = self.interpreter.runcode(command)
            else:
                old_stdout = sys.stdout
                old_stderr = sys.stderr
                sys.stdout = io.StringIO()
                sys.stderr = io.StringIO()

                ret = self.interpreter.runcode(command)

                stdout_output = sys.stdout.getvalue()
                stderr_output = sys.stderr.getvalue()
                sys.stdout = old_stdout
                sys.stderr = old_stderr

                if stdout_output:
                    self.append_output(stdout_output)
                if stderr_output:
                    self.append_output(stderr_output)

                if ret:
                    self.append_output(str(ret))

        except Exception as e:
            self.append_output(str(e))

    def append_output(self, text: str):
        """
        Add some text to the output
        :param text: text to append
        """
        self.console.append_output(text)

    def run_source_code(self):
        """
        Run the source code in the IPython console
        """
        source_code = self.ui.sourceCodeTextEdit.toPlainText()

        if source_code[-1] != '\n':
            source_code += "\n"

        self.console.execute(source_code)
        self.console.append_output(">>> ")

    def source_code_tree_clicked(self, index):
        """
        On double click on a source code tree item, load the source code
        """
        pth = self.python_fs_model.filePath(index)

        if os.path.exists(pth):
            with open(pth, 'r') as f:
                txt = "\n".join(line.rstrip() for line in f)
                self.ui.sourceCodeTextEdit.setPlainText(txt)

            name = os.path.basename(pth)
            self.ui.sourceCodeNameLineEdit.setText(name.replace('.py', ''))
        else:
            error_msg(pth + ' does not exists :/', 'Open script')

    def clear_source_code(self):
        """
        Clear source code
        """
        ok = yes_no_question(text='Are you sure you want to clear source code?',
                             title='Clear Source Code')

        if ok:
            self.ui.sourceCodeNameLineEdit.setText("")
            self.ui.sourceCodeTextEdit.setPlainText("")

    def save_source_code(self):
        """
        Save the source code
        """
        name = self.ui.sourceCodeNameLineEdit.text().strip()

        if name != '':
            fname = name + '.py'
            pth = os.path.join(scripts_path(), fname)
            with open(pth, 'w') as f:
                f.write(self.ui.sourceCodeTextEdit.toPlainText())
        else:
            error_msg("Please enter a name for the script", title="Save script")

    def delete_source_code(self):
        """
        Delete the selected file
        """
        index = self.ui.sourceCodeTreeView.currentIndex()
        pth = self.python_fs_model.filePath(index)
        if os.path.exists(pth):
            ok = yes_no_question(text="Do you want to delete_with_dialogue {}?".format(pth), title="Delete source code file")

            if ok:
                os.remove(pth)
        else:
            error_msg(pth + ' does not exists :/', "Delete source code file")

    def show_source_code_tree_context_menu(self, pos: QtCore.QPoint):
        """
        Show source code tree view context menu
        :param pos: Relative click position
        """
        context_menu = QtWidgets.QMenu(parent=self.ui.diagramsListView)

        gf.add_menu_entry(menu=context_menu,
                          text="Delete",
                          icon_path=":/Icons/icons/delete_with_dialogue.svg",
                          function_ptr=self.delete_source_code)

        # Convert global position to local position of the list widget
        mapped_pos = self.ui.sourceCodeTreeView.viewport().mapToGlobal(pos)
        context_menu.exec(mapped_pos)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations

import os
from PySide6 import QtWidgets
from PySide6.QtCore import Qt

from typing import Dict
from VeraGrid.Gui.GridMerge.build_diff_tree import populate_tree
from VeraGrid.Gui.GridMerge.grid_diff_gui import Ui_Dialog
from VeraGrid.Gui.general_dialogues import LogsDialogue
from VeraGrid.Gui.messages import error_msg
import VeraGrid.Session.file_handler as filedrv

from VeraGridEngine.basic_structures import Logger
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.Devices.types import ALL_DEV_TYPES


def handle_item_changed(item: QtWidgets.QTreeWidgetItem, column: int):
    """

    :param item:
    :param column: THis must be here because this is an event handler
    :return:
    """
    if item.parent() is None:  # Root item
        item.setCheckState(0, Qt.CheckState.Checked)  # always checked...
        for i in range(item.childCount()):
            item.child(i).setCheckState(0, Qt.CheckState.Checked)


class GridDiffDialogue(QtWidgets.QDialog):
    """
    GridMergeDialogue
    """

    def __init__(self, grid: MultiCircuit):
        """

        :param grid:
        """
        QtWidgets.QDialog.__init__(self)
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Grid differential')

        self.ui.treeWidget.setHeaderLabels(
            ["Grid", "Object type", "action", "idtag", "name", "property", "value", "new value"]
        )

        self.logger = Logger()

        self._new_grid: MultiCircuit = grid

        self.all_elms_base_dict: Dict[str, ALL_DEV_TYPES] = {}
        _, ok = self._new_grid.get_all_elements_dict(logger=self.logger)

        if not ok:
            dlg = LogsDialogue('The circuit has duplicated idtags and cannot be differentiated :(', self.logger)
            dlg.exec()
            return

        self._diff: MultiCircuit | None = None

        self.added_grid: bool = False
        self.merged_grid: bool = False
        self.ui.addButton.setText("Open base")
        self.ui.acceptButton.setText("Save diff")

        self.open_file_thread_object = None
        self.save_file_thread_object = None

        # tree item changes
        self.ui.treeWidget.itemChanged.connect(handle_item_changed)
        self.ui.acceptButton.clicked.connect(self.save_diff)
        self.ui.addButton.clicked.connect(self.open_base_grid)

    def open_base_grid(self):
        """
        Open base grid
        :return:
        """
        files_types = "VeraGrid (*.gridcal, *.veragrid)"

        filename, type_selected = QtWidgets.QFileDialog.getOpenFileName(parent=self,
                                                                        caption="Open base grid",
                                                                        filter=files_types)

        if len(filename) > 0:
            if os.path.exists(filename):
                # create thread
                self.open_file_thread_object = filedrv.FileOpenThread(
                    file_name=filename,
                    previous_circuit=None,
                    options=filedrv.FileOpenOptions()
                )

                # make connections
                # self.open_file_thread_object.progress_signal.connect(self.ui.progressBar.setValue)
                # self.open_file_thread_object.progress_text.connect(self.ui.progress_label.setText)
                self.open_file_thread_object.done_signal.connect(self.post_open_base_grid)

                # thread start
                self.open_file_thread_object.start()
            else:
                error_msg(title="File not found", text=f"{filename} not found :(")

        else:
            self.open_file_thread_object = None

    def post_open_base_grid(self):
        """
        After open, make the diff
        :return:
        """
        if self.open_file_thread_object is not None:

            if self.open_file_thread_object.valid:

                # assign the loaded circuit
                if self.open_file_thread_object.circuit is not None:
                    base_grid: MultiCircuit = self.open_file_thread_object.circuit

                    # Create the base dictionary
                    self.all_elms_base_dict, _ = base_grid.get_all_elements_dict(logger=self.logger)

                    ok, logger, self._diff = self._new_grid.differentiate_circuits(base_grid=base_grid,
                                                                                   detailed_profile_comparison=True)

                    if ok:
                        self.build_tree()
                    else:
                        dlg = LogsDialogue('Errors while computing the differential :(', logger)
                        dlg.exec()

    def build_tree(self):
        """
        Build the display tree
        """
        all_no_action = populate_tree(tree_widget=self.ui.treeWidget,
                                      base=self._new_grid,
                                      diff=self._diff,
                                      all_elms_base_dict=self.all_elms_base_dict)

    def save_diff(self):
        """
        Apply the changes
        :return:
        """
        if self._diff is None:
            error_msg(text="No differential created :(\nDid you load a base grid to compare?",
                      title="No diff")
        else:
            # select the file to save
            filename, type_selected = QtWidgets.QFileDialog.getSaveFileName(self, 'Save file',
                                                                            self._diff.name,
                                                                            "VeraGrid diff (*.dveragrid)")

            if filename != '':

                # if the user did not enter the extension, add it automatically
                name, file_extension = os.path.splitext(filename)

                if file_extension == '':
                    filename = name + ".dveragrid"

                self.save_file_thread_object = filedrv.FileSaveThread(circuit=self._diff,
                                                                      file_name=filename,
                                                                      options=filedrv.FileSavingOptions())

                # make connections
                # self.save_file_thread_object.progress_signal.connect(self.ui.progressBar.setValue)
                # self.save_file_thread_object.progress_text.connect(self.ui.progress_label.setText)
                # self.save_file_thread_object.done_signal.connect(self.UNLOCK)
                self.save_file_thread_object.done_signal.connect(self.post_save_diff)

                # thread start
                self.save_file_thread_object.start()

    def post_save_diff(self):
        """

        :return:
        """
        self.close()